/*
 * SmartGWT Mobile
 * Copyright 2008 and beyond, Isomorphic Software, Inc.
 *
 * SmartGWT Mobile is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License version 3
 * as published by the Free Software Foundation.  SmartGWT Mobile is also
 * available under typical commercial license terms - see
 * http://smartclient.com/license
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 */

package com.smartgwt.mobile.client.data;

import java.util.HashMap;
import java.util.Map;

public class DataSourceField {

    private Map<String, Object> attributes = new HashMap<String, Object>();

    public DataSourceField() {
    }

    DataSourceField(Record record) {
        attributes.putAll(record);
    }

    public DataSourceField(String fieldName) {
        attributes.put("name", fieldName);
    }

    public DataSourceField(String fieldName, String type) {
        attributes.put("name", fieldName);
        attributes.put("type", type);
    }

    public DataSourceField(DataSourceField field) {
        attributes.putAll(field.attributes);
    }

    public static DataSourceField combineFieldData(DataSourceField localField, DataSourceField dsField) {
        if (dsField == null && localField == null) {
            return null;
        } else if (dsField == null) {
            return new DataSourceField(localField);
        } else if (localField == null) {
            return new DataSourceField(dsField);
        }

        DataSourceField retField = new DataSourceField(localField);
        Map<String, Object> attributes = retField.attributes;
        for (Map.Entry<String, Object> entry : dsField.attributes.entrySet()) {
            String key = entry.getKey();
            if (attributes.get(key) == null) {
                attributes.put(key, entry.getValue());
            }
        }

        // (shallow) merge nested fields
        Map<String, DataSourceField> localNestedFields = localField.getNestedFields(),
            dsNestedFields = dsField.getNestedFields();
        if (localNestedFields == null && dsNestedFields == null) {
            retField.setNestedFields(null);
        } else {
            Map<String, DataSourceField> retNestedFields = new HashMap<String, DataSourceField>();
            if (dsNestedFields != null) {
                retNestedFields.putAll(dsNestedFields);
            }
            if (localNestedFields != null) {
                retNestedFields.putAll(localNestedFields);
            }
            retField.setNestedFields(retNestedFields);
        }

        retField.setName(localField.getName());
        return retField;
    }

    public boolean getBoolean(String key) {
        Object obj = attributes.get(key);
        if (obj == null) return false;
        if (obj instanceof Boolean) return (Boolean)obj;
        return (obj.equals("true"));
    }

    public void setName(String name) {
        attributes.put("name", name);
    }

    public String getName() {
        return (String)attributes.get("name");
    }

    /**
     * Indicates that this field should always be array-valued (Java type {@link java.util.List}).
     * If the value is singular, it will be wrapped in an array.
     * 
     * <p>Specifically for XML serialization and deserialization, <code>multiple:true</code>
     * behaves similarly to the <a href="http://www.zvon.org/xxl/soapReference/Output/type_array.html">SOAP array idiom</a>;
     * that is, there is a "wrapper element" having the field name as its tag name and one
     * or more child "value" elements that specify the values of the array.
     * 
     * <p>The {@link #getChildTagName() childTagName} attribute allows the tag name of the
     * child "value" elements to be customized.
     */
    public Boolean isMultiple() {
        return (Boolean)attributes.get("multiple");
    }

    public void setMultiple(Boolean multiple) {
        attributes.put("multiple", multiple);
    }

    /**
     * Indicates that this field should have {@link com.smartgwt.mobile.client.data.RecordList} or
     * {@link com.smartgwt.mobile.client.data.Record} values, depending on whether this field
     * is {@link #isMultiple() multiple:true} or not, respectively.  This can be used when
     * records are nested within records.
     * 
     * <p>If <code>multiple</code> is <code>true</code>, {@link #getChildTagName() childTagName}
     * is still applicable.  If not <code>null</code>, as in Example 3 below, <code>childTagName</code>
     * is the name of subelements that will be parsed as records using <code>nestedFields</code>.
     * Otherwise, as in Example 2 below, subelements of the record element having tag name
     * {@link #getName() name} are collected and processed as the subelements.
     * 
     * <table border="1" cellpadding="5">
     *   <caption>Example data source responses and corresponding <code>DataSourceField</code> configurations</caption>
     *   <tr>
     *     <th>&nbsp;</th>
     *     <th>Example 1</th>
     *     <th>Example 2</th>
     *     <th>Example 3</th>
     *   </tr>
     *   <tr>
     *     <th>Data source response</th>
     *     <td>
     *       <pre>&lt;response&gt;
     *  &lt;status&gt;0&lt;/status&gt;
     *  &lt;data&gt;
     *    &lt;record&gt;
     *      &lt;nestedRecord&gt;
     *        &lt;id&gt;7&lt;/id&gt;
     *        &lt;title&gt;Nest #7&lt;/title&gt;
     *      &lt;/nestedRecord&gt;
     *    &lt;/record&gt;
     *  &lt;/data&gt;
     *&lt;/response&gt;</pre>
     *     </td>
     *     <td>
     *       <pre>&lt;response&gt;
     *  &lt;status&gt;0&lt;/status&gt;
     *  &lt;data&gt;
     *    &lt;record&gt;
     *      &lt;nestedRecord id="7"/&gt;
     *      &lt;nestedRecord id="91"/&gt;
     *      &lt;nestedRecord id="12"/&gt;
     *      &lt;nestedRecord id="2"/&gt;
     *    &lt;/record&gt;
     *  &lt;/data&gt;
     *&lt;/response&gt;</pre>
     *     </td>
     *     <td>
     *       <pre>&lt;response&gt;
     *  &lt;status&gt;0&lt;/status&gt;
     *  &lt;data&gt;
     *    &lt;record&gt;
     *      &lt;nestedRecords&gt;
     *        &lt;nestedRecord id="7"/&gt;
     *        &lt;nestedRecord id="91"/&gt;
     *        &lt;nestedRecord id="12"/&gt;
     *        &lt;nestedRecord id="2"/&gt;
     *      &lt;/nestedRecords&gt;
     *    &lt;/record&gt;
     *  &lt;/data&gt;
     *&lt;/response&gt;</pre>
     *     </td>
     *   </tr>
     *   <tr>
     *     <th>{@link #getName() name}</th>
     *     <td><code>"nestedRecord"</code></td>
     *     <td><code>"nestedRecord"</code></td>
     *     <td><code>"nestedRecords"</code></td>
     *   </tr>
     *   <tr>
     *     <th>{@link #isMultiple() multiple}</th>
     *     <td><code>false</code></td>
     *     <td><code>true</code></td>
     *     <td><code>true</code></td>
     *   </tr>
     *   <tr>
     *     <th>{@link #getChildTagName() childTagName}</th>
     *     <td>N/A</td>
     *     <td><code>null</code></td>
     *     <td><code>"nestedRecord"</code></td>
     *   </tr>
     * </table>
     * 
     * @see #isMultiple()
     * @see #getChildTagName()
     * @return a map of the data source fields of the nested records.
     */
    public Map<String, DataSourceField> getNestedFields() {
        @SuppressWarnings("unchecked")
        final Map<String, DataSourceField> nestedFields = (Map<String, DataSourceField>)attributes.get("nestedFields");
        return nestedFields;
    }

    public void setNestedFields(Map<String, DataSourceField> nestedFields) {
        attributes.put("nestedFields", nestedFields);
    }

    /**
     * For a field that is multiple:true, the tag name of each subelement.
     * 
     * <p>If <code>null</code>, then the child tag name is presumed to be "value".
     * 
     * @return String the tag name of subelements.
     */
    public String getChildTagName() {
        return (String)attributes.get("childTagName");
    }
    public void setChildTagName(String childTagName) {
        attributes.put("childTagName", childTagName);
    }

    public void setDerived(boolean derived) {
        attributes.put("derived", derived);
    }

    public boolean isDerived() {
        return getBoolean("derived");
    }

    public void setHidden(boolean hidden) {
        attributes.put("hidden", hidden);
    }

    public boolean isHidden() {
        return getBoolean("hidden");
    }

    public void setRequired(boolean required) {
        attributes.put("required", required);
    }

    public boolean isRequired() {
        return getBoolean("required");
    }

    public void setType(String type) {
        attributes.put("type", type);
    }

    /**
     * The name of a registered {@link com.smartgwt.mobile.client.data.SimpleType}.
     * 
     * @return the name of a registered SimpleType representing the type of this field.
     */
    public String getType() {
        return (String)attributes.get("type");
    }

    public void setLength(Long length) {
        attributes.put("length", length);
    }

    public Long getLength() {
        Object obj = attributes.get("length");
        if (obj == null) return null;
        return Long.parseLong(obj.toString());
    }

    public void setTitle(String title) {
        attributes.put("title", title);
    }

    public String getTitle() {
        return (String)attributes.get("title");
    }

    public void setPrimaryKey(Boolean pk) {
        attributes.put("primaryKey", pk);
    }

    public Boolean isPrimaryKey() {
        return (Boolean)attributes.get("primaryKey");
    }

    public void setForeignKey(String fk) {
        attributes.put("foreignKey", fk);
    }

    public String getForeignKey() {
        return (String)attributes.get("foreignKey");
    }

    public void setInapplicable(Boolean inapplicable) {
        attributes.put("inapplicable", inapplicable);
    }

    public Boolean isInapplicable()  {
        return (Boolean)attributes.get("inapplicable");
    }
}
