package com.smartgwt.mobile.client.data;

import java.util.Date;
import java.util.HashMap;
import java.util.Map;

import com.google.gwt.i18n.client.DateTimeFormat;
import com.google.gwt.i18n.shared.TimeZone;
import com.google.gwt.xml.client.Element;
import com.google.gwt.xml.client.Node;
import com.smartgwt.mobile.client.internal.data.CanFormatDateTime;

public abstract class SimpleType {
    private static final Map<String, SimpleType> registeredTypes = new HashMap<String, SimpleType>();

    private static class NumberSimpleType extends SimpleType {
        public NumberSimpleType(String name) {
            super(name);
        }

        // So as to support the widest range of integers possible, we parse each numeric value
        // in a data source response as a double. This is to match JSONUtils and Smart GWT
        // behavior, where the browser's native JSON-parsing capability is used to parse
        // "integer", "sequence", and "float"-typed values. The loss of precision, if any,
        // occurs at conversion time (e.g. Record#getValueAsInt()).
        //
        // In JavaScript, all numbers are 64-bit IEEE 754 doubles (ECMA-262, Section 8.5),
        // allowing for all integers in the range -2^53 to 2^53 (-9007199254740992 to 9007199254740992), inclusive,
        // to be exactly represented.
        @Override
        public Double parseInput(String text) {
            if (text == null) return null;
            text = text.trim();
            if (text.length() == 0) return null;
            if (text.charAt(0) == '+') {
                text = text.substring(1);
                if (text.length() == 0) return null;
            }
            try {
                return Double.valueOf(text);
            } catch (NumberFormatException ex) {
                return null;
            }
        }
    }

    public static final SimpleType TEXT_TYPE = new SimpleType("text") {
        @Override
        public String parseInput(String text) {
            if (text == null) return null;
            return text.trim();
        }
    };
    
    public static final SimpleType BOOLEAN_TYPE = new SimpleType("boolean") {
        @Override
        public Boolean parseInput(String text) {
            if (text == null) return null;
            text = text.trim();
            if ("true".equalsIgnoreCase(text) || "1".equals(text)) return Boolean.TRUE;
            else if ("false".equalsIgnoreCase(text) || "0".equals(text)) return Boolean.FALSE;
            else return null;
        }
    };

    public static final DateTimeFormat DATE_FORMAT = DateTimeFormat.getFormat("yyyy-MM-dd"),
            DATETIME_FORMAT = DateTimeFormat.getFormat("yyyy-MM-dd'T'HH:mm:ss"),
            TIME_FORMAT = DateTimeFormat.getFormat("HH:mm:ss");

    private static class DateTimeSimpleType extends SimpleType implements CanFormatDateTime {

        private final DateTimeFormat externalFormat, format;

        public DateTimeSimpleType(String name, DateTimeFormat externalFormat) {
            super(name);
            this.externalFormat = externalFormat;
            this.format = DateTimeFormat.getFormat(externalFormat.getPattern() + ".SSS Z");
        }

        @Override
        public Date parseInput(String text) {
            if (text == null) return null;
            try {
                return format.parseStrict(text.trim() + ".000 UT");
            } catch (IllegalArgumentException ex) {
                return null;
            }
        }

        @Override
        public String format(Date date, TimeZone timeZone) {
            return externalFormat.format(date, timeZone);
        }
    }

    public static final SimpleType DATETIME_TYPE = new DateTimeSimpleType("datetime", DATETIME_FORMAT);

    static {
        @SuppressWarnings("unused")
        final SimpleType integerType = new NumberSimpleType("integer");
        @SuppressWarnings("unused")
        final SimpleType sequenceType = new NumberSimpleType("sequence");
        @SuppressWarnings("unused")
        final SimpleType intType = new NumberSimpleType("int");
        @SuppressWarnings("unused")
        final SimpleType floatType = new NumberSimpleType("float");
        @SuppressWarnings("unused")
        final SimpleType doubleType = new NumberSimpleType("double");
        @SuppressWarnings("unused")
        final SimpleType dateType = new SimpleType("date") {

            private final DateTimeFormat format = DateTimeFormat.getFormat(DATE_FORMAT.getPattern() + "'T'HH:mm:ss.SSS");

            @Override
            public Date parseInput(String text) {
                if (text == null) return null;
                try {
                    return format.parseStrict(text.trim() + "T12:00:00.000");
                } catch (IllegalArgumentException ex) {
                    return null;
                }
            }
        };
        @SuppressWarnings("unused")
        final SimpleType timeType = new SimpleType("time") {

            private final DateTimeFormat format = DateTimeFormat.getFormat(TIME_FORMAT.getPattern() + ".SSS");

            @Override
            public Date parseInput(String text) {
                if (text == null) return null;
                try {
                    return format.parseStrict(text.trim() + ".000");
                } catch (IllegalArgumentException ex) {
                    return null;
                }
            }
        };
    }

    public static SimpleType getType(String name) {
        if (name == null) return null;
        return registeredTypes.get(name.toLowerCase());
    }

    private String name;

    public SimpleType(String name) {
        name = name.toLowerCase();
        if (registeredTypes.containsKey(name)) {
            throw new IllegalArgumentException("SimpleType '" + name + "' already registered");
        }
        this.name = name;
        registeredTypes.put(name, this);
    }

    public String getName() {
        return name;
    }

    public Object fromNode(Node node) {
        if (node == null) return null;
        String text = null;
        // For certain node types, nodeValue is null:
        // https://developer.mozilla.org/en/DOM/Node.nodeValue
        if (node instanceof Element) {
            if (node.hasChildNodes()) {
                text = node.getFirstChild().getNodeValue();
            }
        } else text = node.getNodeValue();
        return parseInput(text);
    }

    /**
     * Converts the input string to a value.
     * 
     * <p>This function should not throw an exception.  If an input string cannot be converted
     * to a value, then an implementation should return <code>null</code> or some type-specific
     * default value.
     * 
     * @param text the input string to parse.
     * @return the value represented by <code>text</code>, or <code>null</code> or a type-specific
     * default value if it could not be parsed.
     */
    public abstract Object parseInput(String text);
}
