package com.smartgwt.mobile.client.internal;

import com.google.gwt.core.client.JsArray;
import com.google.gwt.dom.client.Document;
import com.google.gwt.dom.client.Element;
import com.google.gwt.dom.client.InputElement;
import com.google.gwt.dom.client.NativeEvent;
import com.google.gwt.dom.client.Touch;
import com.google.gwt.event.dom.client.TouchEvent;
import com.google.gwt.user.client.Event;
import com.google.gwt.user.client.Event.NativePreviewEvent;
import com.google.gwt.user.client.Event.NativePreviewHandler;
import com.google.gwt.user.client.Timer;
import com.smartgwt.mobile.SGWTInternal;
import com.smartgwt.mobile.client.internal.util.ElementUtil;
import com.smartgwt.mobile.client.widgets.Canvas;

@SGWTInternal
public class EventHandler {

    static {
        captureEvents();
    }

    public static final String LONG_TOUCH_OKAY_CLASS_NAME = "sc-long-touch-okay";

    static void captureEvents() {
        Event.addNativePreviewHandler(new NativePreviewHandler() {

            private boolean trackingClick = false;
            private int clientX, clientY;
            private Long startTime = null;
            private Timer longTouchTimer = null;

            private boolean isLongTouchOkay(Element element) {
                for (; element != null; element = element.getParentElement()) {
                    if (ElementUtil.hasClassName(element, LONG_TOUCH_OKAY_CLASS_NAME)) return true;
                }
                return false;
            }

            // On iOS, it is important to ignore the browser's events on text-type <input> and <textarea> elements
            // because programmatically-created MouseEvents dispatched to those types of input elements
            // do NOT cause the virtual keyboard to pop open and do not give the input element focus.
            // This means that it is impossible to avoid the ~200ms delay between user tapping on a text input
            // and the input receiving keyboard focus.
            // See http://www.quora.com/Mobile-Safari-iPhone-or-iPad-with-JavaScript-how-can-I-launch-the-on-screen-keyboard for
            // background information.
            //
            // Programmatic click events are known to work on button and checkbox INPUTs.
            // It is likely that they also work on radio, reset, and submit INPUTs, but this
            // theory has not been tested yet.
            private boolean shouldIgnore(Element targetElem) {
                String inputType;
                return targetElem == null ||

                       // In HTML documents, element tag names exposed through a JS property are in uppercase.
                       // http://www.w3.org/TR/DOM-Level-2-HTML/html.html#ID-822762427
                       ("INPUT".equals(targetElem.getTagName()) &&
                        !"button".equals(inputType = targetElem.<InputElement>cast().getType()) &&
                        !"checkbox".equals(inputType)) ||
                       "TEXTAREA".equals(targetElem.getTagName());
            }

            private void cancelTracking() {
                if (longTouchTimer != null) {
                    longTouchTimer.cancel();
                    longTouchTimer = null;
                }
                trackingClick = false;
            }

            private void onTouchStart(NativePreviewEvent previewEvent) {
                final NativeEvent event = previewEvent.getNativeEvent();
                final JsArray<Touch> touches = event.getTouches();
                final int clientX, clientY;
                if (touches != null) {
                    if (touches.length() != 1) {
                        cancelTracking();
                        return;
                    }
                    final Touch touch = touches.get(0);
                    clientX = touch.getClientX();
                    clientY = touch.getClientY();
                } else {
                    clientX = event.getClientX();
                    clientY = event.getClientY();
                }
                this.clientX = clientX;
                this.clientY = clientY;

                final Element targetElem = EventUtil.getTargetElem(event);
                if (isLongTouchOkay(targetElem)) {
                    startTime = null;
                } else {
                    startTime = System.currentTimeMillis();
                }
                trackingClick = true;
                if (targetElem != null) {
                    longTouchTimer = new Timer() {

                        @Override
                        public void run() {
                            if (longTouchTimer == this) {
                                longTouchTimer = null;
                                if (trackingClick) {
                                    // http://stackoverflow.com/questions/433919/javascript-simulate-right-click-through-code
                                    final NativeEvent contextMenuEvent = Document.get().createMouseEvent("contextmenu", true, true, 1, 0, 0, clientX, clientY, false, false, false, false, NativeEvent.BUTTON_RIGHT, null);
                                    targetElem.dispatchEvent(contextMenuEvent);
                                }
                            }
                        }
                    };
                    longTouchTimer.schedule(250);
                }

                if (TouchEvent.isSupported()) {
                    // Mobile browsers can give <input> and <textarea> elements keyboard focus if they
                    // are moved in the way. If the target is not a text-type <input> or <textarea> element, then
                    // call preventDefault() on the event.
                    // See: http://stackoverflow.com/questions/11064680/is-this-a-bug-tapping-an-element-can-result-in-a-different-element-receiving-keyboard-focus
                    if (!shouldIgnore(targetElem)) {
                        // TODO Is there a work-around to the issue of wrong text inputs receiving
                        // keyboard focus that does not require calling preventDefault()?
                        //
                        // preventDefault() has the side effect of disabling native scrolling on iOS.
                        if (!Canvas.useIOSNativeScrolling()) {
                            event.preventDefault();
                        }
                    }
                }
            }

            private void onTouchMove(NativePreviewEvent previewEvent) {
                if (! trackingClick) return;

                if (startTime != null && System.currentTimeMillis() - startTime.longValue() > 500L) {
                    cancelTracking();
                    return;
                }

                NativeEvent event = previewEvent.getNativeEvent();
                final JsArray<Touch> touches = event.getTouches();
                final int clientX, clientY;
                if (touches != null) {
                    if (touches.length() != 1) {
                        cancelTracking();
                        return;
                    }
                    final Touch touch = touches.get(0);
                    clientX = touch.getClientX();
                    clientY = touch.getClientY();
                } else {
                    clientX = event.getClientX();
                    clientY = event.getClientY();
                }

                final int deltaX = clientX - this.clientX,
                        deltaY = clientY - this.clientY;
                if (10 < Math.abs(deltaX) || 10 < Math.abs(deltaY)) {
                    cancelTracking();
                    return;
                }
            }

            private native void markFastClick(NativeEvent event) /*-{
                event._fastClick = true;
            }-*/;

            private void onTouchEnd(NativePreviewEvent previewEvent) {
                if (! trackingClick) return;
                cancelTracking();

                Long timeDiff = null;
                if (startTime != null) {
                    timeDiff = Long.valueOf(System.currentTimeMillis() - startTime.longValue());
                    if (timeDiff.longValue() > 500L) return;
                }

                NativeEvent event = previewEvent.getNativeEvent();
                final JsArray<Touch> touches = event.getChangedTouches();
                final int clientX, clientY;
                if (touches != null) {
                    if (touches.length() != 1) return;
                    final Touch touch = touches.get(0);
                    clientX = touch.getClientX();
                    clientY = touch.getClientY();
                } else {
                    clientX = event.getClientX();
                    clientY = event.getClientY();
                }

                int deltaX = clientX - this.clientX,
                        deltaY = clientY - this.clientY;
                if (10 < Math.abs(deltaX) || 10 < Math.abs(deltaY)) {
                    return;
                }

                final Element targetElem = EventUtil.getTargetElem(event);
                if (!shouldIgnore(targetElem)) {
                    if (timeDiff == null || timeDiff.longValue() < 300L) {
                        final NativeEvent clickEvent = Document.get().createClickEvent(1, 0, 0, this.clientX, this.clientY, false, false, false, false);
                        markFastClick(clickEvent);
                        new Timer() {
                            @Override
                            public void run() {
                                targetElem.dispatchEvent(clickEvent);
                            }
                        }.schedule(1);
                    }
                }
            }

            private native boolean isFastClickEvent(NativeEvent event) /*-{
                return !! event._fastClick;
            }-*/;

            @Override
            public void onPreviewNativeEvent(NativePreviewEvent previewEvent) {
                if (previewEvent == null) return;
                final int eventType = previewEvent.getTypeInt();
                if (TouchEvent.isSupported()) {
                    switch (eventType) {
                        case Event.ONTOUCHSTART:
                            onTouchStart(previewEvent);
                            return;
                        case Event.ONTOUCHMOVE:
                            onTouchMove(previewEvent);
                            return;
                        case Event.ONTOUCHEND:
                            onTouchEnd(previewEvent);
                            return;
                        case Event.ONTOUCHCANCEL:
                            trackingClick = false;
                            return;
                    }
                } else {
                    switch (eventType) {
                        case Event.ONMOUSEDOWN:
                            onTouchStart(previewEvent);
                            return;
                        case Event.ONMOUSEMOVE:
                            onTouchMove(previewEvent);
                            return;
                        case Event.ONMOUSEUP:
                            onTouchEnd(previewEvent);
                            return;
                    }
                }
                if (eventType == Event.ONCLICK) {
                    if (!isFastClickEvent(previewEvent.getNativeEvent())) {
                        previewEvent.cancel();
                        return;
                    }
                }
            }
        });
    }

    public static void maybeInit() {}
}

