/*
 * SmartGWT Mobile
 * Copyright 2008 and beyond, Isomorphic Software, Inc.
 *
 * SmartGWT Mobile is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License version 3
 * as published by the Free Software Foundation.  SmartGWT Mobile is also
 * available under typical commercial license terms - see
 * http://smartclient.com/license
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 */

package com.smartgwt.mobile.client.json;

import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import com.google.gwt.i18n.client.DateTimeFormat;
import com.google.gwt.i18n.client.TimeZone;
import com.smartgwt.mobile.client.types.ValueEnum;

public class JSONSerializer {

    private static final int indentSize = 4;
    private static final String spaces = "                                                            ";
    private static final int maxLineLength = 80;

    private static final DateTimeFormat dateTimeFormat = DateTimeFormat.getFormat("yyyy-MM-dd'T'HH:mm:ss");
    private static final TimeZone UTC = TimeZone.createTimeZone(0);

    private static boolean prettyPrint = true;

    public static String serialize(Object in, boolean strict) {
        return serialize(in, null, strict);
    }

    public static String serialize(Object in, String out, boolean strict) {
        if (out == null) out = "";
        return serialize(in, out, 0, strict);
    }

    public static String serialize(Object in, String out, int indent, boolean strict) {
        if (in == null) {
            // Passed a null value to serialize
            return " ";
        }

        if (in instanceof Map) {
            return serializeMap((Map<?, ?>)in, out, indent, strict);
        }

        if (in instanceof List) {
            return serializeList((List<?>)in, out, indent, strict);
        }

        return out;
    }

    private static String quotedForm(CharSequence in, boolean strict) {
        assert in != null;
        StringBuilder sb = new StringBuilder();
        sb.append('"');
        for (int j = 0; j < in.length(); j++) {
            final char c = in.charAt(j);
            if (strict) {
                if (('\u0000' <= c && c <= '\u001F') ||
                    ('\u007F' <= c && c <= '\u009F'))
                {
                    // Control characters
                    // http://www.fileformat.info/info/unicode/category/Cc/list.htm
                    if (c == '\b') sb.append("\\b");
                    else if (c == '\t') sb.append("\\t");
                    else if (c == '\n') sb.append("\\n");
                    else if (c == '\f') sb.append("\\f");
                    else if (c == '\r') sb.append("\\r");
                    else {
                        String hexRep = Integer.toString((int)c, 16);
                        if (hexRep.length() == 1) sb.append("\\u000");
                        else {
                            assert hexRep.length() == 2;
                            sb.append("\\u00");
                        }
                        sb.append(hexRep);
                    }
                } else if (c == '"') sb.append("\\\"");
                else if (c == '\\') sb.append("\\\\");
                else sb.append(c);
            } else {
                // ISC RESTHandler does not support the \ uXXXX syntax in string literals for
                // specifying a Unicode character by code point.
                if (c == '\b') sb.append("\\b");
                else if (c == '\t') sb.append("\\t");
                else if (c == '\n') sb.append("\\n");
                else if (c == '\f') sb.append("\\f");
                else if (c == '\r') sb.append("\\r");
                else if (c == '"') sb.append("\\\"");
                else if (c == '\\') sb.append("\\\\");
                else sb.append(c);
            }
        }
        sb.append('"');
        return sb.toString();
    }

    private static String quotedForm(Object in, boolean strict) {
        if (in == null) return "null";
        else if (in instanceof CharSequence) return quotedForm((CharSequence)in, strict);
        else if (in instanceof Date) { // Special serialization for Dates
            StringBuilder datesb = new StringBuilder();
            datesb.append("\"$$DATE$$:");
            // We really need to consult schema here to decide whether to serialize this as a
            // date or a datetime.  For now, serialize out the time info regardless and let the
            // server do what it likes; SGWT Mobile is missing date/time functionality on the
            // client side, so at the moment this is kind of academic
            datesb.append(dateTimeFormat.format((Date)in, UTC));
            datesb.append("\"");
            return new String(datesb);
        } else if (in instanceof ValueEnum) return quotedForm(((ValueEnum)in).getValue(), strict);
        else if (in instanceof Boolean || in instanceof Number) return in.toString();
        return quotedForm(in.toString(), strict);
    }

    private static boolean isSerializable(Object value) {
        if (value == null ||
            value instanceof CharSequence ||
            value instanceof Enum ||
            value instanceof Boolean ||
            value instanceof Number ||
            value instanceof Date ||
            value instanceof Character)
        {
            return true;
        }
        return false;
    }

    public static String serializeMap(Map<?, ?> map, String out, int indent, boolean strict) {
        if (map == null) return "null";

        int currLineLength = 0;

        if (prettyPrint) out += spaces.substring(0, indent);
        out += "{ ";
        if (prettyPrint) {
            out += "\n" + spaces.substring(0, indent + indentSize);
            currLineLength = indent + indentSize;
        }

        boolean writtenOne = false;
        for (Map.Entry<?, ?> e : map.entrySet()) {
            final String keyStr = quotedForm("" + e.getKey(), strict);
            final Object value = e.getValue();

            if (isSerializable(value)) {
                if (writtenOne) {
                    out += ", ";
                    currLineLength += 2;
                }
                String thisProperty = keyStr + ": " + quotedForm(value, strict);
                if (prettyPrint && currLineLength + thisProperty.length() > maxLineLength) {
                    out += "\n" + spaces.substring(0, indent + indentSize);
                    currLineLength = indent + indentSize;
                }
                out += thisProperty;
                writtenOne = true;
                currLineLength += thisProperty.length();
            } else {
                String innerJS = "";
                innerJS = serialize(value, innerJS, indent + indentSize, strict);
                if (innerJS.equals("")) {
                    // Couldn't serialize property
                } else {
                    if (writtenOne) {
                        out += ", ";
                        currLineLength += 2;
                    }
                    if (prettyPrint) out += "\n" + spaces.substring(0, indent + indentSize);
                    out += keyStr + ": ";
                    out += innerJS; //.substring(indent + indentSize);
                    currLineLength = indent + indentSize + 2;
                    writtenOne = true;
                }
            }
        }

        if (prettyPrint) out += "\n";
        out += "}";
        return out;
    }

    public static String serializeList(List<?> list, String out, int indent, boolean strict) {
        boolean writtenOne = false;
        for (Iterator<?> i = list.iterator(); i.hasNext(); ) {
            Object object = i.next();
            String innerJS = "";
            innerJS = serialize(object, innerJS, indent + indentSize, strict);
            if (!innerJS.equals("")) {
                if (!writtenOne) {
                    if (prettyPrint) out+= spaces.substring(0, indent);
                    out += "[";
                    if (prettyPrint) out+= "\n";
                } else {
                    out += ", ";
                    if (prettyPrint) out+= "\n";
                }
                out += innerJS;
                writtenOne = true;
            }
        }
        
        if (writtenOne) {
            if (prettyPrint) out += "\n" + spaces.substring(0, indent);
            out += "]";
        }
        return out;
    }
}
