/*
 * SmartGWT Mobile
 * Copyright 2008 and beyond, Isomorphic Software, Inc.
 *
 * SmartGWT Mobile is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License version 3
 * as published by the Free Software Foundation.  SmartGWT Mobile is also
 * available under typical commercial license terms - see
 * http://smartclient.com/license
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 */

package com.smartgwt.mobile.client.widgets;

import com.google.gwt.dom.client.DivElement;
import com.google.gwt.dom.client.Document;
import com.google.gwt.dom.client.Element;
import com.google.gwt.dom.client.ImageElement;
import com.google.gwt.dom.client.SpanElement;
import com.google.gwt.dom.client.Style;
import com.google.gwt.dom.client.Style.Display;
import com.google.gwt.dom.client.Style.Unit;
import com.google.gwt.dom.client.StyleInjector;
import com.google.gwt.event.shared.HandlerRegistration;
import com.google.gwt.resources.client.ImageResource;
import com.google.gwt.user.client.Event;
import com.google.gwt.user.client.ui.Image;
import com.smartgwt.mobile.SGWTInternal;
import com.smartgwt.mobile.client.internal.EventUtil;
import com.smartgwt.mobile.client.internal.util.ElementUtil;
import com.smartgwt.mobile.client.internal.widgets.events.ContentChangedEvent;
import com.smartgwt.mobile.client.internal.widgets.events.ContentChangedHandler;
import com.smartgwt.mobile.client.internal.widgets.events.HasContentChangedHandlers;
import com.smartgwt.mobile.client.theme.ThemeResources;
import com.smartgwt.mobile.client.types.IconAlign;
import com.smartgwt.mobile.client.widgets.icons.IconResources;

/**
 * Base abstract class for button types.
 */
public abstract class BaseButton extends Canvas implements HasContentChangedHandlers {

    static {
        StyleInjector.injectAtEnd(ThemeResources.INSTANCE.buttonsCSS().getText());
    }

    public static enum ButtonType {
        BORDERED("bordered"),
        BORDERED_IMPORTANT("bordered important"),
        BORDERED_WARNING("bordered warning"),
        PLAIN("plain-icon"),
        /**
         * A rounded rectangle style button.
         */
        ROUNDED_RECTANGLE("rounded-rectangle"),

        /**
         * A rounded rectangle style button. Used to denote buttons that are more "important"
         */
        ROUNDED_RECTANGLE_IMPORTANT("rounded-rectangle important"),

        /**
         * A default "action" button.
         */
        ACTION_DEFAULT("action default"),

        /**
         * Action button used for "cancel" events.
         */
        ACTION_CANCEL("action cancel"),

        /**
         * Action button used for "delete" events.
         */
        ACTION_DELETE("action delete"),

        /**
         * Action button used for "important" events.
         */
        ACTION_IMPORTANT("action important"),

        /**
         * A "green" action button.
         */
        ACTION_GREEN("action green");

        private String value;

        ButtonType(String value) {
            this.value = value;
        }

        public String getValue() {
            return value;
        }
    }

    @SGWTInternal
    public static void _maybeInit() {}

    public native void changed(BaseButton item,Element input) /*-{
		input.addEventListener('webkitTransitionEnd',function(event){
			item.@com.smartgwt.mobile.client.widgets.BaseButton::_onEnd(Lcom/google/gwt/user/client/Event;)(event);
		},false);			
	}-*/;

    private HandlerRegistration iconLoadRegistration;
	/**
     * The icon for the button
     */
    private Image icon;
    private int top, bottom, left, right;
    protected boolean masked;

    /**
     * The alignment of the icon with respect to its title
     */
    protected IconAlign iconAlign = IconAlign.LEFT;

    /**
     * The color of the icon can be set programmatically. The result will be the icon being displayed
     * in solid "iconColor"
     */
    protected String iconColor;

    /**
     * Buttons support being tinted dynamically.
     */
    protected String tintColor;

    /**
     * The underlying secondary button CSS class name
     */
    private String buttonTypeClassName;


    /**
     * Reference to the "content" element in the button DOM structure
     */
    private DivElement contentElem;

    /**
     * The button title element reference
     */
    private SpanElement labelElem;

    protected boolean stretch = false;
    @SGWTInternal
    protected boolean _touchedActive = false;

    /**
     * Various CSS class names for states / style
     */
    private String baseClassName = "sc-button";
    private String touchedClassName = "touched";
    private String iconOnlyClassName = "icon";
    private String iconMaskClassName = "icon-mask";
    @SGWTInternal
    protected String _tintClassName = "sc-custom-tint";

    /**
     * The default constructor that sets up the basic button structure. A button should have a
     * {@link #setTitle(String) title} and / or an {@link #setIcon(com.google.gwt.user.client.ui.Image, boolean) icon}
     * to have a meaningful appearance.
     */
    public BaseButton() {
        createElements();

        sinkEvents(Event.ONCLICK);
    }

    /**
     * BaseButton constructor.
     *
     * @param title the title label of the button
     */
    public BaseButton(String title) {
        this();
        setTitle(title);
    }

    public BaseButton(BaseButton button) {
        this.baseClassName = button.baseClassName;
        this.touchedClassName = button.touchedClassName;
        this.iconOnlyClassName = button.iconOnlyClassName;
        this.iconMaskClassName = button.iconMaskClassName;
        this._tintClassName = button._tintClassName;
        createElements();

        setButtonTypeClassName(button.buttonTypeClassName);

        setTitle(button.getTitle(), false);
        Image buttonIcon = button.getIcon();
        if (buttonIcon != null) setIcon(buttonIcon, button.masked, false);
        setIconAlign(button.getIconAlign());
        setIconColor(button.getIconColor());
        setTintColor(button.getTintColor());
        setStretch(button.stretch);

        sinkEvents(Event.ONCLICK);
    }

    private void createElements() {
        DivElement buttonElem = Document.get().createDivElement();
        buttonElem.addClassName(baseClassName);
        buttonElem.addClassName(iconOnlyClassName);
        contentElem = Document.get().createDivElement();
        contentElem.addClassName("content");
        labelElem = Document.get().createSpanElement();
        labelElem.addClassName("label");
        contentElem.appendChild(labelElem);
        buttonElem.appendChild(contentElem);
        setElement(buttonElem);
        sinkEvents(Event.ONMOUSEDOWN | Event.ONMOUSEUP | Event.ONCLICK | Event.ONMOUSEOUT | Event.TOUCHEVENTS);
    }
    
    private void markPosition(Event event) {
        top = getElement().getAbsoluteTop();
        left = getElement().getAbsoluteLeft();
        bottom = getElement().getAbsoluteBottom();
        right = getElement().getAbsoluteRight();        
    }
    
    private boolean shouldUnhighlight(Event event) {
        int x = (event.getTouches() != null && event.getTouches().length() > 0) ? event.getTouches().get(0).getClientX(): event.getClientX();
        int y = (event.getTouches() != null && event.getTouches().length() > 0) ? event.getTouches().get(0).getClientY(): event.getClientY();
        return ((x < getElement().getAbsoluteLeft() || x > getElement().getAbsoluteRight()) || 
        (y < getElement().getAbsoluteTop() || y > getElement().getAbsoluteBottom())) || (top != getElement().getAbsoluteTop() || 
        left != getElement().getAbsoluteLeft());
    }

    @Override
    public void destroy() {
        if (iconLoadRegistration != null) {
            iconLoadRegistration.removeHandler();
        }
    }

    /**
     * Sets the title of the button.
     * 
     * @deprecated Use {@link #setTitle(String)} instead.
     */
    @Override
    @Deprecated
    public void setContents(String html) {
        // Note: The reason why this method does not set the innerHTML of `contentElem` is
        // that it is impossible to get the "actual" (unclipped) width of the contents in
        // _getContentWidth().  This is because `contentElem` is display:block.
        // display:inline-block would work, but then text-overflow:ellipsis on the label
        // element no longer works.  Also, overflow:hidden would have to be added to
        // getElement(), but this can change the appearance of the button considerably
        // (for example, NavigationButton places the arrows of "back" and "next" buttons
        // outside of the border box and these are clipped by overflow:hidden on getElement()).
        setTitle(html);
    }

    @Override
    public void setDisabled(boolean disabled) {
        super.setDisabled(disabled);
        if (disabled) onEnd();
    }

    @SGWTInternal
    public final float _getContentWidth() {
        float ret = labelElem.getOffsetWidth();
        if (icon != null && icon.isVisible()) {
            ret += ElementUtil.getOuterWidth(icon.getElement(), true);
        }
        return ret;
    }

    @SGWTInternal
    public final SpanElement _getLabelElement() {
        return labelElem;
    }

    /**
     * Return the title label.
     *
     * @return the title label
     */
    public final String getTitle() {
        return labelElem.getInnerHTML();
    }

    /**
     * Sets the title label of the button.
     *
     * @param title the button title
     */
    public void setTitle(String title) {
        setTitle(title, true);
    }

    private void setTitle(String title, boolean fireContentChangedEvent) {
        labelElem.setInnerHTML(title);
        if (title != null && !title.equals("")) {
            getElement().removeClassName(iconOnlyClassName);
        } else {
            getElement().addClassName(iconOnlyClassName);
        }
        if (fireContentChangedEvent) ContentChangedEvent.fire(this, this);
    }

    /**
     * Return the button icon, or null if not set.
     *
     * @return the button icon
     */
    public final Image getIcon() {
        return icon;
    }

    /**
     * Set the button icon from an ImageResource.
     *
     * @param iconResource the icon image resource.
     * @param mask         true to mask the button. Masking is typically applied to black-only icons causing its final
     *                     appearance to be white
     * @param iconAlign    the alignment of the icon
     */
    public void setIcon(ImageResource iconResource, boolean mask, IconAlign iconAlign) {
        setIcon(new Image(iconResource), mask);
        setIconAlign(iconAlign);
    }

    /**
     * Set the button icon from an ImageResource.
     *
     * @param iconResource the icon image resource.
     * @param mask         true to mask the button. Masking is typically applied to black-only icons causing its final
     *                     appearance to be white
     */
    public void setIcon(ImageResource iconResource, boolean mask) {
        setIcon(new Image(iconResource), mask);
    }

    /**
     * Set the button icon.
     *
     * @param icon the icon image
     * @param mask true to mask the button. Masking is typically applied to black-only icons causing its final
     *             appearance to be white
     */
    public void setIcon(Image icon, boolean mask) {
        setIcon(icon, mask, true);
    }

    private void setIcon(final Image icon, boolean mask, boolean fireContentChangedEvent) {
        if (iconLoadRegistration != null) {
            iconLoadRegistration.removeHandler();
        }
        if (this.icon != null) {
            remove(this.icon);
        }
        this.icon = icon;
        this.masked = mask;

        if (icon != null) {
            if (icon.getWidth() > 0) {
                doSetIcon(false);
            } else {
                icon.getElement().getStyle().setDisplay(Display.NONE);
                iconLoadRegistration = icon.addLoadHandler(new com.google.gwt.event.dom.client.LoadHandler() {

                    @Override
                    public void onLoad(com.google.gwt.event.dom.client.LoadEvent event) {
                        doSetIcon(true);
                        iconLoadRegistration.removeHandler();
                    }
                });
            }
            insert(icon, contentElem.<com.google.gwt.user.client.Element>cast(), 0, true);
        }
        if (fireContentChangedEvent) ContentChangedEvent.fire(this, this);
    }

    private void doSetIcon(boolean fireContentChangedEvent) {
        ImageElement img = ImageElement.as(icon.getElement());
        final Style imgStyle = img.getStyle();
        if (masked) {
            img.setClassName(iconMaskClassName);
            img.setSrc(IconResources.INSTANCE.blank().getSafeUri().asString());
            imgStyle.clearBackgroundImage();
            imgStyle.clearBackgroundColor();
            imgStyle.setProperty("webkitMaskBoxImage", "url(" + icon.getUrl() + ")");
        } else {
            img.removeClassName(iconMaskClassName);
        }
        img.setWidth(23);
        imgStyle.setWidth(23, Unit.PX);
        img.setHeight(23);
        imgStyle.setHeight(23, Unit.PX);

        setIconAlign(iconAlign);
        if (iconColor != null) {
            setIconColor(iconColor);
        }

        imgStyle.clearDisplay();
        if (fireContentChangedEvent) ContentChangedEvent.fire(this, this);
    }

    /**
     * Return the icon alignment.
     *
     * @return the icon alignment. Defaults to left
     */
    public IconAlign getIconAlign() {
        return iconAlign;
    }

    /**
     * Set the icon alignment. Default is left.
     *
     * @param iconAlign the icon alignment
     */
    public void setIconAlign(IconAlign iconAlign) {
        if (this.iconAlign != null) {
            getElement().removeClassName(iconAlign.getValue());
        }
        this.iconAlign = iconAlign;
        if (this.icon != null) {
            getElement().addClassName(iconAlign.getValue());
        }
    }

    /**
     * Return the tint color. Default is null and the theme's default color is used.
     *
     * @return the tint color
     */
    public String getTintColor() {
        return tintColor;
    }

    /**
     * Set the button tintColor. Can pass in the color name, the hext value, or the rgb / rgba value as a String.
     * <br><br>
     * <b>Note:</b> Passing a rgba string with a value for opacity allows buttons to have an opacity / translucency.
     *
     * @param tintColor the tint color
     */
    public void setTintColor(String tintColor) {
        this.tintColor = tintColor;
        if (tintColor != null) {
            getElement().addClassName(_tintClassName);
            getElement().getStyle().setBackgroundColor(tintColor);
        } else {
            getElement().removeClassName(_tintClassName);
            getElement().getStyle().clearBackgroundColor();
        }
    }

    /**
     * Set the icon color. This overrides the default color of the icon an applies a solid color to the icon
     * based on the icon color provided. An example use of this is to set the icon color of a "star" icon to yellow
     * to indicate it has been marked.
     *
     * @param iconColor the icon color. Defaults to null
     */
    public void setIconColor(String iconColor) {
        this.iconColor = iconColor;
        if (this.icon != null) {
            ImageElement imgElement = (ImageElement) contentElem.getChild(0).cast();
            imgElement.getStyle().setBackgroundColor(iconColor);
        }
    }

    /**
     * Return the icon color.
     *
     * @return the icon color. Defaults to null
     */
    public String getIconColor() {
        return iconColor;
    }

      
    /**
     * Should the button stretch in width?
     *
     * @return true for stretch, false for auto-width based on contents
     */
    public boolean isStretch() {
        return stretch;
    }

    /**
     * Should the button stretch in width?
     *
     * @param stretch true for stretch, false for auto-width based on contents
     */
    public void setStretch(boolean stretch) {
        this.stretch = stretch;
        if (stretch) {
            getElement().addClassName("sc-stretch");
        } else {
            getElement().removeClassName("sc-stretch");
        }
    }

    @Override
    public void onBrowserEvent(final Event event) {
        super.onBrowserEvent(event);
        if (! isEnabled()) {
            return;
        }
        Element targetElem = EventUtil.getTargetElem(event);
        if (targetElem != null && getElement().isOrHasChild(targetElem)) {
            final int type = event.getTypeInt();
            switch (type) {
                case Event.ONCLICK:
                    final Element activeElement = ElementUtil.getActiveElement();
                    if (activeElement != null) activeElement.blur();
                    break;
                case Event.ONMOUSEDOWN:
                    if(!touched) {
                        _onStart(event);
                    }
                    break;
                case Event.ONTOUCHSTART:
                    touched = true;
                    _onStart(event);
                    break;
                case Event.ONTOUCHMOVE:
                    _onMove(event);
                    break;
                case Event.ONMOUSEUP:
                    if(!touched) {
                        _onEnd(event);
                    }
                    break;
                case Event.ONTOUCHEND:
                case Event.ONTOUCHCANCEL:
                    touched = true;
                    _onEnd(event);
                    break;
                case Event.ONMOUSEOUT:
                    _onEnd(event);
                    break;
                default:
                    break;
            }
        }
    }

    @SGWTInternal
    void _onStart(Event event) {
        markPosition(event);
        if(!_touchedActive) {
            _touchedActive = true;
            getElement().addClassName(touchedClassName);
        }
    }

    @SGWTInternal
    void _onEnd(Event event) {
        onEnd();
    }
    
    @SGWTInternal
    void _onMove(Event event) {
        if(shouldUnhighlight(event)) {
            _onEnd(event);
        }
    }

    private void onEnd() {
        if(_touchedActive) {
            _touchedActive = false;
            getElement().removeClassName(touchedClassName);
        }
    }

    /**
     * Apply a secondary CSS class name that is specific to each subclass/
     *
     * @param buttonTypeClassName the buttonType CSS class name
     */
    protected void setButtonTypeClassName(String buttonTypeClassName) {
        assert buttonTypeClassName != null : "buttonTypeClassName cannot be null";
        if (this.buttonTypeClassName != null) {
            getElement().removeClassName(this.buttonTypeClassName);
        }
        this.buttonTypeClassName = buttonTypeClassName;
        getElement().addClassName(buttonTypeClassName);
    }

    @Override
    public HandlerRegistration _addContentChangedHandler(ContentChangedHandler handler) {
        return addHandler(handler, ContentChangedEvent.getType());
    }
}
