/*
 * SmartGWT Mobile
 * Copyright 2008 and beyond, Isomorphic Software, Inc.
 *
 * SmartGWT Mobile is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License version 3
 * as published by the Free Software Foundation.  SmartGWT Mobile is also
 * available under typical commercial license terms - see
 * http://smartclient.com/license
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 */
package com.smartgwt.mobile.client.widgets;

import java.util.HashMap;
import java.util.List;

import com.google.gwt.event.shared.HandlerRegistration;
import com.smartgwt.mobile.SGWTInternal;
import com.smartgwt.mobile.client.data.Criteria;
import com.smartgwt.mobile.client.data.DataSource;
import com.smartgwt.mobile.client.data.Record;
import com.smartgwt.mobile.client.data.ResultSet;
import com.smartgwt.mobile.client.data.SortSpecifier;
import com.smartgwt.mobile.client.data.events.HasDataChangedHandlers;
import com.smartgwt.mobile.client.internal.RecordSelection;
import com.smartgwt.mobile.client.internal.Selection;
import com.smartgwt.mobile.client.internal.data.HasCriteria;
import com.smartgwt.mobile.client.internal.data.HasRecordCanSelectPropertyAttribute;
import com.smartgwt.mobile.client.internal.data.HasSortSpecifiers;
import com.smartgwt.mobile.client.types.FetchMode;
import com.smartgwt.mobile.client.types.TextMatchStyle;
import com.smartgwt.mobile.client.widgets.grid.events.HasSelectionUpdatedHandlers;
import com.smartgwt.mobile.client.widgets.grid.events.SelectionUpdatedEvent;
import com.smartgwt.mobile.client.widgets.grid.events.SelectionUpdatedHandler;
import com.smartgwt.mobile.client.widgets.layout.VLayout;

public abstract class DataBoundComponent<D extends List</*? extends */Record> & HasDataChangedHandlers> extends VLayout implements HasSelectionUpdatedHandlers, HasSortSpecifiers {

    private static final int DEFAULT_DATA_PAGE_SIZE = 25;

    private FetchMode dataFetchMode = FetchMode.PAGED;
    private int dataPageSize = DEFAULT_DATA_PAGE_SIZE;
    private DataSource dataSource;
    private Criteria initialCriteria = null;
    private D data;
    private String titleField;
    private String infoField = "info";
    private String dataField;
    private String descriptionField = "description";
    private String iconField = "icon";
    @SGWTInternal
    protected SortSpecifier[] _sortSpecifiers;

    private Selection<Record> selection;

    public final FetchMode getDataFetchMode() {
        return dataFetchMode;
    }

    public void setDataFetchMode(FetchMode fetchMode) {
        this.dataFetchMode = fetchMode;
        if (data != null && data instanceof ResultSet) {
            ((ResultSet)data).setFetchMode(fetchMode);
        }
    }

    public final int getDataPageSize() {
        return dataPageSize;
    }

    public void setDataPageSize(Integer dataPageSize) {
        this.dataPageSize = dataPageSize == null ? DEFAULT_DATA_PAGE_SIZE : dataPageSize.intValue();
    }

    @SGWTInternal
    protected void _createSelectionModel() {
        if (selection != null) _destroySelectionModel();

        Selection<Record> selection = new RecordSelection(data);
        if (this instanceof HasRecordCanSelectPropertyAttribute) {
            selection.setCanSelectProperty(((HasRecordCanSelectPropertyAttribute)this).getRecordCanSelectProperty());
        }

        this.selection = selection;
    }

    @SGWTInternal
    protected void _destroySelectionModel() {
        if (selection == null) return;
        selection._destroy();
        selection = null;
    }

    public void setDataSource(DataSource dataSource) {
        this.dataSource = dataSource;
    }

    public final DataSource getDataSource() {
        return dataSource;
    }

    @SGWTInternal
    public Criteria _getInitialCriteria() {
        return initialCriteria;
    }

    public final Criteria getCriteria() {
        if (data == null || data.size() == 0) return new Criteria(initialCriteria);
        else if (data instanceof HasCriteria) return new Criteria(((HasCriteria)data).getCriteria());
        else return null;
    }

    public void setCriteria(Criteria criteria) {
        if (data instanceof HasCriteria) ((HasCriteria)data).setCriteria(criteria);
        else initialCriteria = criteria;
    }

    @SGWTInternal
    protected final D _getData() {
        return data;
    }

    private static int nextNoNullOrDuplicateRecordsRunNum = 1;
    private static <D extends List</*? extends */Record>> boolean noNullOrDuplicateRecords(D data) {
        if (data != null && !data.isEmpty()) {
            final Integer thisNoNullOrDuplicateRecordsRunNum = Integer.valueOf(nextNoNullOrDuplicateRecordsRunNum++);
            final int data_size = data.size();
            for (int i = 0; i < data_size; ++i) {
                final Record record = data.get(i);
                assert record != null : "The Record at index " + i + " of the given RecordList is null.";
                final Integer noNullOrDuplicateRecordsRunNum = record.getAttributeAsInt("$noNullOrDuplicateRecordsRunNum");
                assert noNullOrDuplicateRecordsRunNum == null ||
                        !noNullOrDuplicateRecordsRunNum.equals(thisNoNullOrDuplicateRecordsRunNum) : "The Record at index " + i + " is contained in the RecordList two or more times (the exact same Record object). Each Record in the RecordList must be a distinct Object.";
                record.setAttribute("$noNullOrDuplicateRecordsRunNum", thisNoNullOrDuplicateRecordsRunNum);
            }
        }
        return true;
    }

    @SGWTInternal
    protected void _setData(D newData) {
        assert noNullOrDuplicateRecords(newData);
        _destroySelectionModel();
        data = newData;
        _createSelectionModel();
    }

    public void setTitleField(String titleField) {
        this.titleField = titleField;
    }

    public final String getTitleField() {
        if (this.titleField != null) {
            return this.titleField;
        }
        if (dataSource != null) {
            return dataSource.getTitleField();
        }
        // Return default name for title attribute.
        return "title";
    }

    public void setInfoField(String infoField) {
        this.infoField = infoField;
    }

    public final String getInfoField() {
        if (infoField != null) {
            return infoField;
        }
        if (dataSource != null) {
            return dataSource.getInfoField();
        }
        return null;
    }

    public void setDataField(String dataField) {
        this.dataField = dataField;
    }

    public final String getDataField() {
        if (dataField != null) {
            return dataField;
        }
        if (dataSource != null) {
            return dataSource.getDataField();
        }
        return null;
    }

    public void setDescriptionField(String descriptionField) {
        this.descriptionField = descriptionField;
    }

    public final String getDescriptionField() {
        if (descriptionField != null) {
            return descriptionField;
        }
        if (dataSource != null) {
            return dataSource.getDescriptionField();
        }
        return null;
    }

    public void setIconField(String iconField) {
        this.iconField = iconField;
    }

    public final String getIconField() {
        if (iconField != null) {
            return iconField;
        }
        if (dataSource != null) {
            return dataSource.getIconField();
        }
        return null;
    }

    public SortSpecifier[] getSort() {
        return SortSpecifier._shallowClone(_sortSpecifiers);
    }

    public void setSort(SortSpecifier... sortSpecifiers) {
        _sortSpecifiers = SortSpecifier._shallowClone(sortSpecifiers);
        if (data != null && sortSpecifiers != null && sortSpecifiers.length > 0) {
            // TODO reset contexts
            if (data instanceof HasSortSpecifiers) ((HasSortSpecifiers)data).setSort(_sortSpecifiers);
            // else if (this.data.sortByProperty) // TODO
        }
    }

    @SGWTInternal
    protected final Selection<Record> _getSelectionObject() {
        return selection;
    }

    public boolean isSelected(Record record) {
        if (record == null || selection == null) return false;
        return selection.isSelected(record);
    }

    public boolean isPartiallySelected(Record record) {
        if (record == null || selection == null) return false;
        return selection.isPartiallySelected(record);
    }

    public final void selectRecord(Record record) {
        selectRecord(record, true);
    }

    public final void selectRecord(Record record, boolean state) {
        selectRecords(new Record[] { record }, state);
    }

    public final void selectSingleRecord(Record record) {
        if (selection == null) return;
        selection.deselectAll();
        selectRecord(record);
    }

    public final void deselectRecord(Record record) {
        selectRecord(record, false);
    }

    public final void selectRecords(Record[] records) {
        selectRecords(records, true);
    }

    public void selectRecords(Record[] records, boolean state) {
        if (selection != null) {
            selection.selectList(records, state);
            _fireSelectionUpdated();
        }
    }

    public final void deselectRecords(Record[] records) {
        selectRecords(records, false);
    }

    public void deselectAllRecords() {
        if (selection != null) {
            selection.deselectAll();
            _fireSelectionUpdated();
        }
    }

    @SGWTInternal
    protected void _fireSelectionUpdated() {
        SelectionUpdatedEvent.fire(this);
    }

    @SGWTInternal
    protected Record[] _getSelection(boolean excludePartialSelections) {
        if (selection == null) return null;
        return selection.getSelection(excludePartialSelections);
    }

    public Record[] getSelectedRecords() {
        return _getSelection(false);
    }

    public Record getSelectedRecord() {
        if (selection == null) return null;
        return selection.getSelectedRecord();
    }

    public abstract void fetchData();

    public void invalidateCache() {
        if (data instanceof ResultSet) {
            ((ResultSet)data).invalidateCache();
        }
    }

    public void filterData() {
        if (dataSource != null) {
            if (dataSource.getRequestProperties() == null) {
                dataSource.setRequestProperties(new HashMap<String, Object>());
            }
            dataSource.getRequestProperties().put("textMatchStyle", TextMatchStyle.SUBSTRING);
            fetchData();            
        }
    }

    @Override
    public HandlerRegistration addSelectionUpdatedHandler(SelectionUpdatedHandler handler) {
        return addHandler(handler, SelectionUpdatedEvent.getType());
    }
}
