/*
 * SmartGWT Mobile
 * Copyright 2008 and beyond, Isomorphic Software, Inc.
 *
 * SmartGWT Mobile is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License version 3
 * as published by the Free Software Foundation.  SmartGWT Mobile is also
 * available under typical commercial license terms - see
 * http://smartclient.com/license
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 */
package com.smartgwt.mobile.client.widgets;

import java.util.ArrayList;
import java.util.List;

import com.google.gwt.core.client.JavaScriptObject;
import com.google.gwt.dom.client.Document;
import com.google.gwt.dom.client.ImageElement;
import com.google.gwt.dom.client.NativeEvent;
import com.google.gwt.dom.client.Style.Display;
import com.google.gwt.dom.client.Style.Unit;
import com.google.gwt.dom.client.StyleInjector;
import com.google.gwt.event.shared.HandlerRegistration;
import com.google.gwt.resources.client.ImageResource;
import com.google.gwt.user.client.Timer;
import com.smartgwt.mobile.SGWTInternal;
import com.smartgwt.mobile.client.internal.widgets.PromptDialog;
import com.smartgwt.mobile.client.theme.ThemeResources;
import com.smartgwt.mobile.client.types.DefaultDialogAction;
import com.smartgwt.mobile.client.widgets.events.ButtonClickEvent;
import com.smartgwt.mobile.client.widgets.events.ButtonClickHandler;
import com.smartgwt.mobile.client.widgets.events.ClickEvent;
import com.smartgwt.mobile.client.widgets.events.ClickHandler;
import com.smartgwt.mobile.client.widgets.events.HasButtonClickHandlers;
import com.smartgwt.mobile.client.widgets.layout.Layout;
import com.smartgwt.mobile.internal.gwt.dom.client.DOMConstants;
import com.smartgwt.mobile.internal.gwt.dom.client.SuperElement;

public class Dialog extends Panel implements HasButtonClickHandlers {

    static {
        StyleInjector.injectAtStart(ThemeResources.INSTANCE.dialogCSS().getText());
    }

    public static final Dialog Prompt = PromptDialog.INSTANCE;

    /**
     * OK  Button object to fire dialog's "okClick()" method on click.
     */
    public static final Button OK = new Button("OK");
    /**
     * APPLY Button object to fire dialog's "applyClick()" method on click.
     */
    public static final Button APPLY = new Button("Apply");
    /**
     * YES Button object to fire dialog's "yesClick()" method on click
     */
    public static final Button YES = new Button("Yes");
    /**
     * NO  Button object to fire dialog's "noClick()" method on click.
     */
    public static final Button NO = new Button("No");
    /**
     * CANCEL  Button object to fire dialog's "cancelClick()" method on click.
     */
    public static final Button CANCEL = new Button("Cancel");
    /**
     * DONE  Button object to fire dialog's "doneClick()" method on click.
     */
    public static final Button DONE = new Button("Done");

    private DefaultDialogAction defaultAction = DefaultDialogAction.HIDE;

    private Header1 titleWidget;

    private Panel contentPanel;
    private Panel iconContainer;
    protected Panel screenCover;

    private Button[] buttons = null;

    private int iconSize = 32;

    private List<HandlerRegistration> handlerRegistrations = new ArrayList<HandlerRegistration>();

    /**
     * Whether the <code>Dialog</code> is in the process of showing itself.
     */
    private boolean isShowing = false;

    /**
     * Whether the <code>Dialog</code> is in the process of hiding itself.
     */
    private boolean isHiding = false;

    private boolean hideAfterShowing = false;

    /**
     * Whether {@link #destroy()} has been called, but this dialog was in transition at the
     * time (either it was showing or hiding itself).  If this dialog is showing itself and
     * <code>shouldDestroy</code>, then it should hide itself.  If this dialog is hiding
     * itself and <code>shouldDestroy</code>, then it should call <code>destroy()</code> once
     * fully hidden.
     */
    private boolean shouldDestroy = false;

    protected Dialog() {
        super(null);
        init();
    }

    public Dialog(Layout parent) {
        super(null);
        init();

        parent.addMember(screenCover);
        parent.addMember(this);
    }

    private native JavaScriptObject createAnimationEndListener() /*-{
        var self = this;
        var listener = $entry(function (event) {
            self.@com.smartgwt.mobile.client.widgets.Dialog::_onAnimationEnd(Lcom/google/gwt/dom/client/NativeEvent;Ljava/lang/String;D)(event, event.animationName, event.elapsedTime);
        });
        return listener;
    }-*/;

    private void init() {
        @SuppressWarnings("unused")
        HandlerRegistration animationEndRegistration = getElement().<SuperElement>cast().addEventListener(DOMConstants.INSTANCE.getAnimationEndEventType(), createAnimationEndListener(), false);

        setClassName("sc-action-sheet");
        getElement().getStyle().setDisplay(Display.NONE);
        if(!(this instanceof Popover)) {
            getElement().getStyle().setBottom(0, Unit.PX);
        }

        titleWidget = new Header1();
        header.addMember(titleWidget);
        header.addStyleName("header");
        addMember(header);
        header.setVisible(false);

        if(!(this instanceof Popover)) {
            contentPanel = new Panel();
            contentPanel.getElement().addClassName("sc-content");
            iconContainer = new Panel();
            iconContainer.getElement().getStyle().setProperty("textAlign", "center");
            addMember(iconContainer);
            addMember(contentPanel);
        }

        screenCover = new Panel();
        screenCover.setClassName("sc-screen-cover");
        screenCover.getElement().getStyle().setDisplay(Display.NONE);
    }

    public Dialog(Layout parent, Button... buttons) {
        this(parent);
        setButtons(buttons);
    }

    public Dialog(Layout parent, String title) {
        this(parent);
        setTitle(title);
    }

    public Dialog(Layout parent, String title, Button... buttons) {
        this(parent);
        setTitle(title);
        setButtons(buttons);
    }

    @SGWTInternal
    protected void _onAnimationEnd(NativeEvent event, String animationName, double elapsedTime) {
        if ("sc-action-sheet_fadeIn".equals(animationName)) {
            _onShown();
        } else if ("sc-action-sheet_fadeOut".equals(animationName)) {
            _onHidden();
        }
    }

    @SGWTInternal
    protected final Panel _getScreenCover() {
        return screenCover;
    }

    @SGWTInternal
    protected final Panel _getContentPanel() {
        return contentPanel;
    }

    /**
     * The standard action that is executed by {@link #doDefaultAction()} when a built-in
     * button (e.g. {@link #YES}, {@link #CANCEL}) on the dialog is clicked.
     * 
     * @return the standard action to perform.  Default value:
     * {@link com.smartgwt.mobile.client.types.DefaultDialogAction#HIDE}.
     */
    public DefaultDialogAction getDefaultAction() {
        return defaultAction;
    }

    public void setDefaultAction(DefaultDialogAction defaultAction) {
        this.defaultAction = defaultAction;
    }

    private void removeHandlerRegistrations() {
        for (HandlerRegistration reg : handlerRegistrations) {
            reg.removeHandler();
        }
        handlerRegistrations.clear();
    }

    @Override
    public void destroy() {
        if (isShowing || isHiding) {
            shouldDestroy = true;
            return;
        }
        if (isVisible()) {
            shouldDestroy = true;
            hide();
            return;
        }
        shouldDestroy = false;
        removeHandlerRegistrations();
        super.destroy();
    }

    /**
     * Title for this Dialog, shown in the header. If the dialog has no title, then the header
     * will not be shown.
     *
     * @param title new title.
     */
    @Override
    public void setTitle(String title) {
        super.setTitle(title);
        if (title != null && title.length() > 0) {
            titleWidget.setContents(title);
            header.setVisible(true);
        } else {
            header.setVisible(false);
        }
    }

    /**
     * Size of the icon to show in this dialog.
     *
     * @param iconSize iconSize Default value is 32
     */
    public void setIconSize(int iconSize) {
        this.iconSize = iconSize;
        if (getIcon() != null) {
            this.setIcon(getIcon());
        }
    }

    /**
     * Size of the icon to show in this dialog.
     *
     * @return int
     */
    public int getIconSize() {
        return iconSize;
    }

    /**
     * Icon to show in this dialog. Icon will be showed above the dialog's message.
     *
     * @param icon icon
     */
    @Override
    public void setIcon(ImageResource icon) {
        super.setIcon(icon);
        if(iconContainer != null) {
            iconContainer.clear();
            if (icon != null) {
                // TODO: maybe we should replace it with Image widget when it will be created
                ImageElement imageElement = Document.get().createImageElement();
                imageElement.setSrc(icon.getURL());
                imageElement.setHeight(getIconSize());
                imageElement.setWidth(getIconSize());
                iconContainer.getElement().appendChild(imageElement);
            }
        }
    }

    /**
     * Message to show in this dialog. <P> If a message is set the primary purpose of the
     * dialog will be assumed to be to show a message with buttons - auto-sizing to the message
     * text will be enabled, and, if 
     * {@link com.smartgwt.mobile.client.widgets.Dialog#setIcon} has also been set, the 
     * {@link com.smartgwt.mobile.client.widgets.Dialog#getMessage} and 
     * {@link com.smartgwt.mobile.client.widgets.Dialog#getIcon}
     * If any of these behaviors are inconvenient or you want more precise control over a
     * message and some custom widgets, start from the superclass
     * {@link com.smartgwt.mobile.client.widgets.Panel} instead, and add controls via
     * {@link com.smartgwt.mobile.client.widgets.Panel#addMember(Canvas) Panel.addMember}.
     *
     * @param message message Default value is null
     */
    public void setMessage(String message) {
        if(contentPanel != null) {
            contentPanel.setContents(message);
        }
    }

    /**
     * Message to show in this dialog. <P> If a message is set the primary purpose of the
     * dialog will be assumed to be to show a message with buttons - auto-sizing to the message
     * text will be enabled, and, if 
     * {@link com.smartgwt.mobile.client.widgets.Dialog#setIcon} has also been set, the 
     * {@link com.smartgwt.mobile.client.widgets.Dialog#getMessage}  
     * If any of these behaviors are inconvenient or you want more precise control over a
     * message and some custom widgets, start from the superclass
     * {@link com.smartgwt.mobile.client.widgets.Panel} instead, and add controls via
     * {@link com.smartgwt.mobile.client.widgets.Panel#addMember(Canvas) Panel.addMember}.
     *
     * @return String
     */
    public String getMessage() {
        if(contentPanel != null) {
            return contentPanel.getInnerHTML();
        }
        return null;
    }

    /**
     * Array of Buttons to show in the Dialog
     * <P>
     * The set of buttons to use is typically set by calling one of the shortcuts such as
     * {@link com.smartgwt.mobile.client.widgets.Dialog#OK OK} or 
     * {@link com.smartgwt.mobile.client.widgets.Dialog#CANCEL CANCEL}. A custom set of
     * buttons can be passed to these shortcuts methods via the "properties" argument, or to a
     * directly created Dialog. 
     * <P>
     * In both cases, a mixture of built-in buttons, custom buttons, and other components can
     * be passed.  Built-in buttons can be referred to as <code>Dialog.OK</code>.
     * <P>
     * If you want to use one of the default button appearance use css styles like 'cancel',
     * 'green' or 'delete'.
     *
     * @param buttons buttons Default value is null
     */
    public void setButtons(Button... buttons) {
        if (this.isVisible()) throw new IllegalStateException();

        removeHandlerRegistrations();

        if (this.buttons != null) {
            for (int i = 0; i < this.buttons.length; ++i) {
                removeMember(this.buttons[i]);
            }
        }

        if (buttons == null) {
            this.buttons = null;
        } else {
            this.buttons = new Button[buttons.length];
            for (int i = 0; i < buttons.length; ++i) {
                final int finalIndex = i;
                Button button = buttons[i];

                if (button == Dialog.OK) {
                    button = new Button(Dialog.OK);
                    final Button finalButton = button;
                    handlerRegistrations.add(button.addClickHandler(new ClickHandler() {
                        @Override
                        public void onClick(ClickEvent event) {
                            okClick();
                            ButtonClickEvent.fire(Dialog.this, finalButton, finalIndex);
                        }
                    }));
                } else if (button == Dialog.APPLY) {
                    button = new Button(Dialog.APPLY);
                    final Button finalButton = button;
                    handlerRegistrations.add(button.addClickHandler(new ClickHandler() {
                        @Override
                        public void onClick(ClickEvent event) {
                            applyClick();
                            ButtonClickEvent.fire(Dialog.this, finalButton, finalIndex);
                        }
                    }));
                } else if (button == Dialog.YES) {
                    button = new Button(Dialog.YES);
                    final Button finalButton = button;
                    handlerRegistrations.add(button.addClickHandler(new ClickHandler() {
                        @Override
                        public void onClick(ClickEvent event) {
                            yesClick();
                            ButtonClickEvent.fire(Dialog.this, finalButton, finalIndex);
                        }
                    }));
                } else if (button == Dialog.NO) {
                    button = new Button(Dialog.NO);
                    final Button finalButton = button;
                    handlerRegistrations.add(button.addClickHandler(new ClickHandler() {
                        @Override
                        public void onClick(ClickEvent event) {
                            noClick();
                            ButtonClickEvent.fire(Dialog.this, finalButton, finalIndex);
                        }
                    }));
                } else if (button == Dialog.CANCEL) {
                    button = new Button(Dialog.CANCEL);
                    final Button finalButton = button;
                    handlerRegistrations.add(button.addClickHandler(new ClickHandler() {
                        @Override
                        public void onClick(ClickEvent event) {
                            cancelClick();
                            ButtonClickEvent.fire(Dialog.this, finalButton, finalIndex);
                        }
                    }));
                } else if (button == Dialog.DONE) {
                    button = new Button(Dialog.DONE);
                    final Button finalButton = button;
                    handlerRegistrations.add(button.addClickHandler(new ClickHandler() {
                        @Override
                        public void onClick(ClickEvent event) {
                            doneClick();
                            ButtonClickEvent.fire(Dialog.this, finalButton, finalIndex);
                        }
                    }));
                } else {
                    final Button finalButton = button;
                    handlerRegistrations.add(button.addClickHandler(new ClickHandler() {
                        @Override
                        public void onClick(ClickEvent event) {
                            ButtonClickEvent.fire(Dialog.this, finalButton, finalIndex);
                        }
                    }));
                }

                addMember(button);
                this.buttons[i] = button;
            }
        }
    }

    @SGWTInternal
    protected final boolean _isHiding() {
        return isHiding;
    }

    /**
     * Makes this dialog invisible.
     */
    public void hide() {
        //SC.logWarn("Dialog.hide(): isHiding = " + isHiding + ", isShowing = " + isShowing + ", isVisible() = " + isVisible());
        if (isHiding || ! isVisible()) return;

        if (isShowing) {
            hideAfterShowing = true;
            return;
        }

        isShowing = false;
        isHiding = true;
        removeClassName("visible");
        setClassName("hidden");
        screenCover.removeClassName("visible");
        screenCover.setClassName("hidden");
        _doHide();
    }

    @SGWTInternal
    protected void _doHide() {}

    @SGWTInternal
    protected void _onHidden() {
        //SC.logWarn("Dialog._onHidden(): isHiding = " + isHiding + ", isShowing = " + isShowing + ", isVisible() = " + isVisible());
        assert isHiding;
        isHiding = false;
        getElement().getStyle().setDisplay(Display.NONE);
        screenCover.getElement().getStyle().setDisplay(Display.NONE);
        if (shouldDestroy) {
            destroy();
        }
    }

    @SGWTInternal
    protected final boolean _getHideAfterShowing() {
        return hideAfterShowing;
    }

    @SGWTInternal
    protected final boolean _isShowing() {
        return isShowing;
    }

    /**
     * Shows dialog at the center of the screen.
     */
    public void show() {
        //SC.logWarn("Dialog.show(): isHiding = " + isHiding + ", isShowing = " + isShowing + ", isVisible() = " + isVisible());
        if (isShowing || isVisible()) return;

        isHiding = false;
        isShowing = true;
        new Timer() {

            @Override
            public void run() {
                screenCover.getElement().getStyle().clearDisplay();
                getElement().getStyle().clearDisplay();
                removeClassName("hidden");
                setClassName("visible");
                screenCover.removeClassName("hidden");
                screenCover.setClassName("visible");
                _doShow();
            }
        }.schedule(1);
    }

    @SGWTInternal
    protected void _doShow() {}

    @SGWTInternal
    protected void _onShown() {
        //SC.logWarn("Dialog._onShown(): isHiding = " + isHiding + ", isShowing = " + isShowing + ", isVisible() = " + isVisible());
        assert isShowing;
        isShowing = false;
        if (shouldDestroy) {
            hide();
        } else if (hideAfterShowing) {
            hide();
            hideAfterShowing = false;
        }
    }

    /**
     * Runs the default action.
     * 
     * @see #getDefaultAction()
     */
    protected void doDefaultAction() {
        switch (defaultAction) {
            case HIDE:
                hide();
                break;
            case DESTROY:
                hide();
                destroy();
                break;
            case DO_NOTHING:
                break;
        }
    }

    /**
     * Called when the {@link #OK} button is clicked.  By default, calls {@link #doDefaultAction()}.
     */
    protected void okClick() {
        doDefaultAction();
    }

    /**
     * Called when the {@link #APPLY} button is clicked.  By default, calls {@link #doDefaultAction()}.
     */
    protected void applyClick() {
        doDefaultAction();
    }

    /**
     * Called when the {@link #YES} button is clicked.  By default, calls {@link #doDefaultAction()}.
     */
    protected void yesClick() {
        doDefaultAction();
    }

    /**
     * Called when the {@link #NO} button is clicked.  By default, calls {@link #doDefaultAction()}.
     */
    protected void noClick() {
        doDefaultAction();
    }

    /**
     * Called when the {@link #CANCEL} button is clicked.  By default, calls {@link #doDefaultAction()}.
     */
    protected void cancelClick() {
        doDefaultAction();
    }

    /**
     * Called when the {@link #DONE} button is clicked.  By default, calls {@link #doDefaultAction()}.
     */
    protected void doneClick() {
        doDefaultAction();
    }

    @Override
    public HandlerRegistration addButtonClickHandler(ButtonClickHandler handler) {
        return addHandler(handler, ButtonClickEvent.getType());
    }
}
