package com.smartgwt.mobile.client.widgets.form;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map;

import com.google.gwt.core.client.GWT;
import com.google.gwt.event.shared.HandlerRegistration;
import com.google.gwt.user.client.Element;
import com.google.gwt.user.client.ui.Widget;
import com.smartgwt.mobile.SGWTInternal;
import com.smartgwt.mobile.client.data.Criteria;
import com.smartgwt.mobile.client.data.Criterion;
import com.smartgwt.mobile.client.data.DSCallback;
import com.smartgwt.mobile.client.data.DSRequest;
import com.smartgwt.mobile.client.data.DSResponse;
import com.smartgwt.mobile.client.data.DataSource;
import com.smartgwt.mobile.client.data.Record;
import com.smartgwt.mobile.client.data.RecordList;
import com.smartgwt.mobile.client.internal.Array;
import com.smartgwt.mobile.client.internal.widgets.form.DynamicFormImpl;
import com.smartgwt.mobile.client.theme.FormCssResource;
import com.smartgwt.mobile.client.theme.ThemeResources;
import com.smartgwt.mobile.client.types.DSOperationType;
import com.smartgwt.mobile.client.types.FormStyle;
import com.smartgwt.mobile.client.util.SC;
import com.smartgwt.mobile.client.widgets.DataBoundComponent;
import com.smartgwt.mobile.client.widgets.form.events.HasItemChangedHandlers;
import com.smartgwt.mobile.client.widgets.form.events.ItemChangedEvent;
import com.smartgwt.mobile.client.widgets.form.events.ItemChangedHandler;
import com.smartgwt.mobile.client.widgets.form.fields.FormItem;
import com.smartgwt.mobile.client.widgets.form.fields.HiddenItem;

public class DynamicForm extends DataBoundComponent<RecordList> implements HasItemChangedHandlers {

    private static final FormCssResource CSS = ThemeResources.INSTANCE.formCSS();
    static {
        CSS.ensureInjected();
    }

    /**
     * Do two field values match?  Used wherever we need to compare field values.
     * Handles all expected data types.
     * Used to detect changes to values (e.g. <code>valuesHaveChanged()</code>).
     */
    @SGWTInternal
    public static boolean _compareValues(Object value1, Object value2) {
        if (value1 == value2) return true;
        if (value1 == null) return value2 == null;
        else if (value2 == null) {
            assert value1 != null;
            return false;
        }
        assert value1 != null && value2 != null;

        Class<?> value1Class = value1.getClass();
        if (value1Class.isArray()) {
            Class<?> value2Class = value2.getClass();
            if (! value2Class.isArray()) return false;
            final int value1Length = Array.arrayLength(value1);
            if (value1Length != Array.arrayLength(value2)) return false;
            for (int i = 0; i < value1Length; ++i) {
                final Object o1 = Array.arrayGet(value1, i),
                        o2 = Array.arrayGet(value2, i);
                if (o1 == null) {
                    if (o2 != null) return false;
                } else if (o2 == null) {
                    assert o1 != null;
                    return false;
                } else {
                    assert o1 != null && o2 != null;
                    if (! o1.equals(o2)) return false;
                }
            }
            return true;
        }

        return value1.equals(value2);
    }

    private final DynamicFormImpl impl = GWT.create(DynamicFormImpl.class);
    private FormStyle formStyle = FormStyle.STYLE1;
    private final Map<String, FormItem> fields = new LinkedHashMap<String, FormItem>();
    private DSOperationType saveOperationType;
    private Record oldValues;

    public DynamicForm() {
        getElement().addClassName(CSS.style1Class());
    }

    @Override
    public void destroy() {
        impl.destroyImpl(this);
        super.destroy();
    }

    /**
     * The display style of this <code>DynamicForm</code>.
     * 
     * <p>{@link com.smartgwt.mobile.client.types.FormStyle#STYLE1 FormStyle.STYLE1} makes the
     * form look like a grouped <code>TableView</code>. {@link com.smartgwt.mobile.client.types.FormStyle#STYLE2 FormStyle.STYLE2}
     * makes the form look like the Edit Info form of the iPhone Contacts app.
     * 
     * @return the display style. Default value: <code>FormStyle.STYLE1</code>
     */
    public final FormStyle getFormStyle() {
        return formStyle;
    }

    public void setFormStyle(FormStyle formStyle) {
        final FormStyle newFormStyle = formStyle == null ? FormStyle.STYLE1 : formStyle;
        if (this.formStyle != newFormStyle) {
            this.formStyle = newFormStyle;
            switch (newFormStyle) {
                case STYLE1:
                    getElement().removeClassName(CSS.style2Class());
                    getElement().addClassName(CSS.style1Class());
                    break;
                case STYLE2:
                    getElement().removeClassName(CSS.style1Class());
                    getElement().addClassName(CSS.style2Class());
                    break;
            }
        }
    }

    @SGWTInternal
    public final Map<String, FormItem> _getFieldsMap() {
        return fields;
    }

    public final FormItem getField(String name) {
        return fields.get(name);
    }

    public final FormItem getItem(String name) {
        return getField(name);
    }

    public final FormItem[] getFields() {
        return fields.values().toArray(new FormItem[fields.size()]);
    }

    public final FormItem[] getItems() {
        return getFields();
    }

    public void setFields(FormItem... items) {
        if (!fields.isEmpty()) {
            for (final FormItem field : fields.values()) {
                field._setForm(null);
                removeMember(field);
            }
            clearErrors();
            fields.clear();
        }
        impl.setFields(this, items);
    }

    public final void setItems(FormItem... items) {
        setFields(items);
    }

    public final DSOperationType getSaveOperationType() {
        return saveOperationType;
    }

    public void setSaveOperationType(DSOperationType saveOperationType) {
        if (saveOperationType == null) saveOperationType = DSOperationType.ADD;
        else if (saveOperationType != DSOperationType.ADD && saveOperationType != DSOperationType.UPDATE) {
            SC.logWarn("Invalid saveOperationType:'" + saveOperationType.getValue() + "'. Assuming 'add'.");
        }
        this.saveOperationType = saveOperationType;
        if (saveOperationType == DSOperationType.ADD) oldValues = null;
    }

    @SGWTInternal
    public final void _add(Widget child, Element container) {
        super.add(child, container);
    }

    @SGWTInternal
    public <T extends Map<String, Object>> T _putValues(T values) {
        if (values != null) {
            for (Map.Entry<String, FormItem> e : fields.entrySet()) {
                values.put(e.getKey(), e.getValue().getValue());
            }
        }
        return values;
    }

    public final Map<String, Object> getValues() {
        return _putValues(new HashMap<String, Object>());
    }

    public final Criteria getValuesAsCriteria() {
       final Criteria criteria = new Criteria();
       assert fields != null;
       for (String key : fields.keySet()) {
           FormItem formItem = (FormItem)fields.get(key);
           Criterion criterion = formItem.getCriterion();
           if (criterion != null) {
               criteria.addCriteria(criterion.getFieldName(), criterion.getValue());
           }
       }
       return criteria;
    }

    public final Record getValuesAsRecord() {
        return _putValues(new Record());
    }

    public void clearErrors() {
        impl.clearErrors(this);
    }

    public void setValues(Map<String, Object> values) {
        assert fields != null;
        // Reset FormItems for which there is no value in `values`.
        for (Iterator<Map.Entry<String, FormItem>> it = fields.entrySet().iterator(); it.hasNext(); /*empty*/) {
            final Map.Entry<String, FormItem> e = it.next();
            if (values == null || !values.containsKey(e.getKey())) {
                final FormItem item = e.getValue();
                if (item instanceof HiddenItem) {
                    item._setForm(null);
                    it.remove();
                } else {
                    item.setValue(null);
                }
            }
        }

        if (values != null) {
            for (Map.Entry<String, Object> e : values.entrySet()) {
                FormItem formItem = fields.get(e.getKey());
                if (formItem != null) formItem.setValue(e.getValue());
                else {
                    formItem = new HiddenItem(e.getKey());
                    formItem.setValue(e.getValue());
                    formItem._setForm(this);
                    fields.put(e.getKey(), formItem);
                }
            }
        }
    }

    public void clearValue(String fieldName) {
        assert fields != null;
        final FormItem item = fields.get(fieldName);
        if (item != null) item.setValue(null);
    }

    public void clearValues() {
        setValues(null);
    }

    public void setErrors(Map<String, Object> errors, boolean showErrors) {
        clearErrors();
        if (errors == null) return;
        for (final String fieldName : errors.keySet()) {
            final FormItem item = fields.get(fieldName);
            if (item != null) {
                String baseStyle = item.getClassName();
                if (baseStyle != null) {
                    item.getElement().addClassName(CSS.formItemValueInvalidClass());
                    final Object errorObj = errors.get(fieldName);
                    if (errorObj != null) {
                        impl.showError(this, showErrors, item, errorObj);
                    }
                }
            }
        }
    }

    public void editRecord(Record record) {
        setValues(record);
        oldValues = getValuesAsRecord();
        setSaveOperationType(DSOperationType.UPDATE);
    }

	public void fetchData() {
		final DataSource dataSource = getDataSource();
		if(dataSource != null) {
			dataSource.fetchData(null, new DSCallback(){
				@Override
				public void execute(DSResponse response, Object rawData, DSRequest request) {
					Record records[] = response.getRecords();
					if(records != null) {
                        FormItem[] items = new FormItem[records.length];
						for(int i = 0; i < records.length; ++i) {
							Record record = records[i];
							String name = (String) record.get("name");
							String type = (String) record.get("type");
                            items[i] =  FormItem._create(name, type);
						}
                        setItems(items);
					}
				}				
			});
		}
		
	}

	public void saveData() {
		saveData(null);
	}

    public void saveData(final DSCallback callback) {
        final DataSource ds = getDataSource();
        if (ds != null) {
            final DSCallback saveCallback = new DSCallback() {
                @Override
                public void execute(DSResponse dsResponse, Object rawData, DSRequest dsRequest) {
                    setErrors(dsResponse.getErrors(), true);
                    if (callback != null) callback.execute(dsResponse, rawData, dsRequest);
                }
            };

            if (saveOperationType == null || saveOperationType == DSOperationType.ADD) {
                ds.addData(getValuesAsRecord(), saveCallback);
            } else {
                assert saveOperationType == DSOperationType.UPDATE;
                final DSRequest requestProps = new DSRequest();
                requestProps.setOldValues(oldValues);
                ds.updateData(getValuesAsRecord(), saveCallback, requestProps);
            }
        }
    }

	public boolean validate() {
		return true;	
	}

    @Override
    public HandlerRegistration addItemChangedHandler(ItemChangedHandler handler) {
        return addHandler(handler, ItemChangedEvent.TYPE);
    }
}
