/*
 * SmartGWT Mobile
 * Copyright 2008 and beyond, Isomorphic Software, Inc.
 *
 * SmartGWT Mobile is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License version 3
 * as published by the Free Software Foundation.  SmartGWT Mobile is also
 * available under typical commercial license terms - see
 * http://smartclient.com/license
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 */

package com.smartgwt.mobile.client.widgets.layout;

import java.util.ArrayList;
import java.util.List;

import com.google.gwt.dom.client.Document;
import com.google.gwt.dom.client.Element;
import com.google.gwt.dom.client.SpanElement;
import com.google.gwt.dom.client.Style;
import com.google.gwt.dom.client.Style.Unit;
import com.google.gwt.dom.client.StyleInjector;
import com.google.gwt.event.logical.shared.ResizeEvent;
import com.google.gwt.event.logical.shared.ResizeHandler;
import com.google.gwt.event.shared.HandlerRegistration;
import com.google.gwt.user.client.Timer;
import com.google.gwt.user.client.Window;
import com.google.gwt.user.client.ui.Widget;
import com.smartgwt.mobile.SGWTInternal;
import com.smartgwt.mobile.client.internal.util.AnimationUtil.Direction;
import com.smartgwt.mobile.client.internal.util.ElementUtil;
import com.smartgwt.mobile.client.internal.widgets.NavigationButtons;
import com.smartgwt.mobile.client.internal.widgets.events.ContentChangedEvent;
import com.smartgwt.mobile.client.internal.widgets.events.ContentChangedHandler;
import com.smartgwt.mobile.client.internal.widgets.layout.NavigationBarItem;
import com.smartgwt.mobile.client.internal.widgets.layout.NavigationItem;
import com.smartgwt.mobile.client.theme.ThemeResources;
import com.smartgwt.mobile.client.widgets.BaseButton;
import com.smartgwt.mobile.client.widgets.Canvas;
import com.smartgwt.mobile.client.widgets.ContainerFeatures;
import com.smartgwt.mobile.client.widgets.Header1;
import com.smartgwt.mobile.client.widgets.NavigationButton;
import com.smartgwt.mobile.client.widgets.NavigationButton.NavigationDirection;
import com.smartgwt.mobile.client.widgets.Panel;
import com.smartgwt.mobile.client.widgets.PanelContainer;

public class NavigationBar extends Canvas implements PanelContainer {

    static {
        BaseButton._maybeInit();

        StyleInjector.injectAtEnd(ThemeResources.INSTANCE.navigationBarCSS().getText());
    }

    @SuppressWarnings("rawtypes")
    private NavStack navStack;
    private final List<NavigationItem> items = new ArrayList<NavigationItem>();

    private HandlerRegistration windowResizeRegistration;
    private HandlerRegistration leftButtonContentChangedRegistration,
            titleViewContentChangedRegistration,
            rightBarItemContentChangedRegistration;

    private Timer layoutTimer;

    private NavigationItem ensureItem() {
        final NavigationItem item;
        final int items_size = items.size();
        if (items_size == 0) {
            item = new NavigationItem();
            items.add(item);
        } else {
            item = items.get(items_size - 1);
        }
        return item;
    }

    NavigationBar(@SuppressWarnings("rawtypes") NavStack stack) {
        setElement(Document.get().createDivElement());
        setClassName("sc-navigationbar");
        getElement().setClassName(this.getClassName());
        this.navStack = stack;
    }

    private void removeHandlers() {
        if (rightBarItemContentChangedRegistration != null) {
            rightBarItemContentChangedRegistration.removeHandler();
            rightBarItemContentChangedRegistration = null;
        }
        if (titleViewContentChangedRegistration != null) {
            titleViewContentChangedRegistration.removeHandler();
            titleViewContentChangedRegistration = null;
        }
        if (leftButtonContentChangedRegistration != null) {
            leftButtonContentChangedRegistration.removeHandler();
            leftButtonContentChangedRegistration = null;
        }
    }

    @Override
    public void destroy() {
        removeHandlers();
        super.destroy();
    }

    @Override
    protected void onLoad() {
        super.onLoad();
        windowResizeRegistration = Window.addResizeHandler(new ResizeHandler() {

            private Timer timer;

            @Override
            public void onResize(ResizeEvent event) {
                // A timeout is used to prevent severe choppiness when the user is resizing
                // a desktop browser window.
                if (timer == null) {
                    timer = new Timer() {
                        @Override
                        public void run() {
                            if (timer != this) return;
                            _layOutMembers();
                            timer = null;
                        }
                    };
                    timer.schedule(200);
                }
            }
        });
        _layOutMembers();
    }

    @Override
    protected void onUnload() {
        windowResizeRegistration.removeHandler();
        super.onUnload();
    }

    @SGWTInternal
    void _clear() {
        removeHandlers();
        if (items != null) {
            while (!items.isEmpty()) {
                final NavigationItem topItem = items.remove(items.size() - 1);
                if (topItem.getLeftButton() != null) {
                    remove(topItem.getLeftButton());
                }
                if (topItem.getTitleView() != null) {
                    remove(topItem.getTitleView());
                }
                if (topItem.getRightBarItem() != null) {
                    remove(topItem.getRightBarItem());
                }
            }
        }
    }

    public final NavigationButton getRightButton() {
        final Canvas rightBarItem = ensureItem().getRightBarItem();
        if (rightBarItem instanceof NavigationButton) {
            return (NavigationButton)rightBarItem;
        } else if (rightBarItem instanceof NavigationButtons) {
            return ((NavigationButtons)rightBarItem).getFirstButton();
        }
        return null;
    }

    @SGWTInternal
    public <T extends Canvas & NavigationBarItem> void _setRightBarItem(T rightBarItem) {
        final NavigationItem item = ensureItem();
        setRightBarItem(item, item, rightBarItem, true, Direction.NONE);
    }

    public void setRightButton(NavigationButton rightButton) {
        _setRightBarItem(rightButton);
    }

    public void setRightButtons(NavigationButton... rightButtons) {
        _setRightBarItem(new NavigationButtons(rightButtons));
    }

    @SGWTInternal
    public <T extends Canvas & NavigationBarItem> void _setRightBarItem(int numSlotsFromTop, T rightBarItem) {
        if (numSlotsFromTop == 0) {
            _setRightBarItem(rightBarItem);
        } else {
            final NavigationItem item = items.get(items.size() - numSlotsFromTop - 1);
            item.setRightBarItem(rightBarItem);
        }
    }

    private <T extends Canvas & NavigationBarItem> void setRightBarItem(NavigationItem oldItem, NavigationItem item, T rightBarItem, boolean doLayOut, Direction direction) {
        Canvas button = oldItem.getRightBarItem();
        if (button == rightBarItem) return;
        if (button != null) {
            deferredRemove(this, button, button.getElement());
            switch(direction) {
            case RIGHT:
                button.setClassName("sc-slide-fadeout-to-left");
                break;
            case LEFT:
                button.setClassName("sc-slide-fadeout-to-right");
                break;
            case UP:
            case DOWN:
            case NONE:
                button.setClassName("sc-fadeout");
                break;
            }
        }
        item.setRightBarItem(rightBarItem);
        if (rightBarItemContentChangedRegistration != null) {
            rightBarItemContentChangedRegistration.removeHandler();
        }
        if (rightBarItem == null) {
            rightBarItemContentChangedRegistration = null;
        } else {
            rightBarItemContentChangedRegistration = rightBarItem._addContentChangedHandler(new ContentChangedHandler() {

                @Override
                public void _onContentChanged(ContentChangedEvent event) {
                    _layOutMembers();
                }
            });
            postAnimate(rightBarItem.getElement());
            if(button != null) {
                switch(direction) {
                case RIGHT:
                    rightBarItem.setClassName("sc-slide-fadein-from-right");
                    break;
                case LEFT:
                    rightBarItem.setClassName("sc-slide-fadein-from-left");
                    break;
                case UP:
                case DOWN:
                case NONE:
                    rightBarItem.setClassName("sc-fadein");
                    break;
                }
            } else {
                rightBarItem.setClassName("sc-fadein");
            }
            add(rightBarItem, getElement().<com.google.gwt.user.client.Element>cast());
        }
        if (doLayOut) _layOutMembers();
    }

    public final NavigationButton getLeftButton() {
        return ensureItem().getLeftButton();
    }

    public void setLeftButton(NavigationButton leftButton) {
        final NavigationItem item = ensureItem();
        setLeftButton(item, item, leftButton, true, Direction.NONE);
    }

    private static native void deferredRemove(NavigationBar bar, Canvas item, Element element) /*-{
        var listener = $entry(function (event) {
            if (element != event.target) return;
            bar.@com.smartgwt.mobile.client.widgets.layout.NavigationBar::remove(Lcom/smartgwt/mobile/client/widgets/Canvas;)(item);
            element.removeEventListener(@com.smartgwt.mobile.internal.gwt.dom.client.DOMConstants::INSTANCE.@com.smartgwt.mobile.internal.gwt.dom.client.DOMConstants::getAnimationEndEventType()(), listener, false);
            listener = null;
            element.className = element.className.replace(/(?:sc-fadeout|sc-slide-fadeout-to-)[a-z]*[ ]?/g, '');
        });
        element.addEventListener(@com.smartgwt.mobile.internal.gwt.dom.client.DOMConstants::INSTANCE.@com.smartgwt.mobile.internal.gwt.dom.client.DOMConstants::getAnimationEndEventType()(), listener, false);
    }-*/;

    private static native void postAnimate(Element element) /*-{
        var listener = $entry(function (event) {
            if (element != event.target) return;
            element.removeEventListener(@com.smartgwt.mobile.internal.gwt.dom.client.DOMConstants::INSTANCE.@com.smartgwt.mobile.internal.gwt.dom.client.DOMConstants::getAnimationEndEventType()(), listener, false);
            listener = null;
            element.className = element.className.replace(/(?:sc-fadein|sc-slide-fadein-from-)[a-z]*[ ]?/g, '');
        });
        element.addEventListener(@com.smartgwt.mobile.internal.gwt.dom.client.DOMConstants::INSTANCE.@com.smartgwt.mobile.internal.gwt.dom.client.DOMConstants::getAnimationEndEventType()(), listener, false);
    }-*/;

    private void remove(Canvas w) {
        super.remove((Widget)w);
    }

    private void insert(Canvas w, Element container, int beforeIndex) {
        super.insert((Widget)w, (com.google.gwt.user.client.Element) container, beforeIndex, true);
    }

    private void setLeftButton(NavigationItem oldItem, NavigationItem item, NavigationButton leftButton, boolean doLayOut, Direction direction) {
        NavigationButton button = oldItem.getLeftButton();
        if (button == leftButton) return;
        if (button != null) {
            deferredRemove(this, button, button.getElement());
            switch(direction) {
            case RIGHT:
                button.setClassName("sc-slide-fadeout-to-left");
                break;
            case LEFT:
                button.setClassName("sc-slide-fadeout-to-right");
                break;
            case UP:
            case DOWN:
            case NONE:
                button.setClassName("sc-fadeout");
                break; 
            }
        }
        item.setLeftButton(leftButton);
        if (leftButtonContentChangedRegistration !=  null) {
            leftButtonContentChangedRegistration.removeHandler();
        }
        if (leftButton == null) {
            leftButtonContentChangedRegistration = null;
        } else {
            leftButtonContentChangedRegistration = leftButton._addContentChangedHandler(new ContentChangedHandler() {

                @Override
                public void _onContentChanged(ContentChangedEvent event) {
                    _layOutMembers();
                }
            });
            postAnimate(leftButton.getElement());
            switch(direction) {
            case RIGHT:
                leftButton.setClassName("sc-slide-fadein-from-right");
                break;
            case LEFT:
                leftButton.setClassName("sc-slide-fadein-from-left");
                break;
            case UP:
            case DOWN:
            case NONE:
                leftButton.setClassName("sc-fadein");
                break;
            }
            insert(leftButton, getElement().<com.google.gwt.user.client.Element>cast(), 0, true);
        }
        if (doLayOut) _layOutMembers();
    }

    public final Header1 getTitleLabel() {
        return ensureItem().getTitleView();
    }

    @SGWTInternal
    public void _layOutMembers() {
        if (! isAttached()) return;
        if (! isVisible() || ! navStack.isVisible()) return;
        if (layoutTimer == null) {
            layoutTimer = new Timer() {

                @Override
                public void run() {
                    final float _1vw = Window.getClientWidth(); // 1 viewport width
                    final NavigationItem item = ensureItem();
                    final Header1 titleView = item.getTitleView();
                    final float minTitleViewWidth, titleViewContentWidth;
                    if (titleView == null) {
                        minTitleViewWidth = 0.0f;
                        titleViewContentWidth = 0.0f;
                    } else {
                        final SpanElement span = titleView._getSpan();
                        span.getStyle().clearPadding();
                        final float titleViewWidth = ElementUtil.getOuterWidth(titleView.getElement(), true);
                        minTitleViewWidth = titleViewWidth - titleView.getElement().getClientWidth();
                        titleViewContentWidth = span.getOffsetWidth();
                    }
                    final float titleViewAutoWidth = minTitleViewWidth + titleViewContentWidth;
                    final NavigationButton leftButton = item.getLeftButton();
                    final Canvas rightBarItem = item.getRightBarItem();
                    final float leftButtonHMargin, absoluteMinLeftButtonWidth, leftButtonContentWidth;
                    if (leftButton == null) {
                        leftButtonHMargin = 0.0f;
                        absoluteMinLeftButtonWidth = 0.0f;
                        leftButtonContentWidth = 0.0f;
                    } else {
                        final Element element = leftButton.getElement();
                        leftButtonHMargin = ElementUtil.getHMarginWidth(element);
                        absoluteMinLeftButtonWidth = ElementUtil.getMinMarginBoxWidth(element);
                        leftButtonContentWidth = leftButton._getContentWidth();
                    }
                    final float rightBarItemHMargin, absoluteMinRightBarItemWidth, rightBarItemContentWidth;
                    if (rightBarItem == null) {
                        rightBarItemHMargin = 0.0f;
                        absoluteMinRightBarItemWidth = 0.0f;
                        rightBarItemContentWidth = 0.0f;
                    } else {
                        final Element element = rightBarItem.getElement();
                        rightBarItemHMargin = ElementUtil.getHMarginWidth(element);
                        absoluteMinRightBarItemWidth = ElementUtil.getMinMarginBoxWidth(element);
                        rightBarItemContentWidth = ((NavigationBarItem)rightBarItem)._getContentWidth();
                    }
                    final float leftButtonAutoWidth = absoluteMinLeftButtonWidth + leftButtonContentWidth,
                            rightBarItemAutoWidth = absoluteMinRightBarItemWidth + rightBarItemContentWidth;

                    final float totalWidth = titleViewAutoWidth + leftButtonAutoWidth + rightBarItemAutoWidth;

                    float newLeftButtonMarginBoxWidth = 0.0f, newRightButtonMarginBoxWidth = 0.0f;
                    if (totalWidth + 0.000001f >= _1vw) {
                        final float minLeftButtonWidth = leftButton == null ? 0.0f : absoluteMinLeftButtonWidth + Math.min(53.0f, leftButtonContentWidth),
                                minRightBarItemWidth = rightBarItem == null ? 0.0f : absoluteMinRightBarItemWidth + (rightBarItem instanceof NavigationButtons ? rightBarItemContentWidth : Math.min(53.0f, rightBarItemContentWidth));
                        assert minLeftButtonWidth <= leftButtonAutoWidth + 0.000001f;
                        assert minRightBarItemWidth <= rightBarItemAutoWidth + 0.000001f;

                        final float r = _1vw - titleViewAutoWidth - minLeftButtonWidth - minRightBarItemWidth;
                        if (r < 0.000001f) {
                            newLeftButtonMarginBoxWidth = minLeftButtonWidth;
                            newRightButtonMarginBoxWidth = minRightBarItemWidth;
                        } else { // The title gets as much width as it needs while still giving the left and/or right
                                 // buttons a minimum client width (at most 53px).  The remaining width is distributed to
                                 // the left and right buttons.
                            if (leftButton != null || rightBarItem != null) {
                                final float denominator = leftButtonAutoWidth - minLeftButtonWidth + rightBarItemAutoWidth - minRightBarItemWidth;
                                if (denominator < 2.0f + 0.000001f) {
                                    newLeftButtonMarginBoxWidth = minLeftButtonWidth;
                                    newRightButtonMarginBoxWidth = minRightBarItemWidth;
                                } else {
                                    newLeftButtonMarginBoxWidth = minLeftButtonWidth + (leftButtonAutoWidth - minLeftButtonWidth) / denominator * r;
                                    newRightButtonMarginBoxWidth = minRightBarItemWidth + (rightBarItemAutoWidth - minRightBarItemWidth) / denominator * r;
                                }
                            }
                        }
                    } else {
                        newLeftButtonMarginBoxWidth = leftButtonAutoWidth;
                        newRightButtonMarginBoxWidth = rightBarItemAutoWidth;

                        final float leftButtonExtra = Math.max(0.0f, leftButtonAutoWidth - rightBarItemAutoWidth),
                                rightButtonExtra = Math.max(0.0f, rightBarItemAutoWidth - leftButtonAutoWidth);
                        float r = _1vw - totalWidth;
                        assert r + 0.000001f >= 0.0f;
                        if (titleView != null) {
                            if (r + 0.000001f >= leftButtonExtra + rightButtonExtra) {
                                final float rightPadding = leftButtonExtra,
                                        leftPadding = rightButtonExtra;
                                final SpanElement span = titleView._getSpan();
                                final Style spanStyle = span.getStyle();
                                spanStyle.setPaddingRight(rightPadding, Unit.PX);
                                spanStyle.setPaddingLeft(leftPadding, Unit.PX);
                            } else {
                                if (leftButtonExtra + 0.000001f >= rightButtonExtra) {
                                    final float rightPadding = r;
                                    final SpanElement span = titleView._getSpan();
                                    span.getStyle().setPaddingRight(rightPadding, Unit.PX);
                                } else {
                                    final float leftPadding = r;
                                    final SpanElement span = titleView._getSpan();
                                    span.getStyle().setPaddingLeft(leftPadding, Unit.PX);
                                }
                            }
                        }
                    }
                    if (leftButton != null) {
                        ElementUtil.setBorderBoxWidth(leftButton.getElement(), newLeftButtonMarginBoxWidth - leftButtonHMargin);
                    }
                    if (rightBarItem != null) {
                        ElementUtil.setBorderBoxWidth(rightBarItem.getElement(), newRightButtonMarginBoxWidth - rightBarItemHMargin);
                    }
                    if (titleView != null) {
                        titleView.getElement().getStyle().setPaddingLeft(newLeftButtonMarginBoxWidth, Unit.PX);
                        titleView.getElement().getStyle().setPaddingRight(newRightButtonMarginBoxWidth, Unit.PX);
                        titleView.getElement().getStyle().clearVisibility();
                    }

                    layoutTimer = null;
                } // run()
            };
            layoutTimer.schedule(1);
        }
    }

    public void setTitleLabel(String titleLabelHtml) {
        setTitleLabel(new Header1(titleLabelHtml));
    }

    public void setTitleLabel(Header1 titleLabel) {
        final NavigationItem item = ensureItem();
        setTitleLabel(item, item, titleLabel, true, Direction.LEFT);
    }

    private void setTitleLabel(NavigationItem oldItem, NavigationItem item, Header1 titleLabel, boolean doLayOut, Direction direction) {
        final Header1 title = oldItem.getTitleView();
        if (title == titleLabel) return;
        if (title != null) {
            deferredRemove(this, title, title.getElement());
            if(direction == Direction.RIGHT) { 
                title.setClassName("sc-slide-fadeout-to-left");
            } else {
                title.setClassName("sc-slide-fadeout-to-right");
            }
        }
        item.setTitleView(titleLabel);
        if (titleViewContentChangedRegistration != null) {
            titleViewContentChangedRegistration.removeHandler();
        }
        if (titleLabel == null) {
            titleViewContentChangedRegistration = null;
        } else {
            final int beforeIndex;
            if (item.getLeftButton() != null) {
                beforeIndex = 1;
            } else {
                beforeIndex = 0;
            }

            titleViewContentChangedRegistration = titleLabel._addContentChangedHandler(new ContentChangedHandler() {

                @Override
                public void _onContentChanged(ContentChangedEvent event) {
                    _layOutMembers();
                }
            });
            if (navStack.size() > 0) {
                postAnimate(titleLabel.getElement());
                if (direction == Direction.RIGHT) {
                    titleLabel.setClassName("sc-slide-fadein-from-right");
                } else {
                    titleLabel.setClassName("sc-slide-fadein-from-left");
                }
            }
            if (title != null) {
                insert(titleLabel, getElement().<com.google.gwt.user.client.Element>cast(), 2, true);
            } else {
                insert(titleLabel, getElement().<com.google.gwt.user.client.Element>cast(), beforeIndex, true);
            }
        }
        if (doLayOut) _layOutMembers();
    }

    <T extends Canvas & NavigationBarItem> void _push(Panel panel) {
        final Panel oldPanel = navStack.top();
        final NavigationItem oldItem = ensureItem();
        final NavigationItem item = new NavigationItem();
        items.add(item);
        final NavigationButton backButton = new NavigationButton(oldPanel.getTitle(), NavigationDirection.BACK);
        setLeftButton(oldItem, item, backButton, false, Direction.RIGHT);
        this.<T>setRightBarItem(oldItem, item, null, false, Direction.RIGHT);
        setTitleLabel(oldItem, item, new Header1(panel.getTitle()), false, Direction.RIGHT);
        _layOutMembers();
    }

    <T extends Canvas & NavigationBarItem> void _pop() {
        final NavigationItem oldItem = items.remove(items.size() - 1);
        NavigationItem item = ensureItem();
        final Panel panel = navStack.top();
        final String newTitle = panel.getTitle();
        final Header1 titleLabel = new Header1(newTitle);
        setTitleLabel(oldItem, item, titleLabel, false, Direction.LEFT);
        setLeftButton(oldItem, item, item.getLeftButton(), false, Direction.LEFT);
        @SuppressWarnings("unchecked")
        T rightBarItem = (T)item.getRightBarItem();
        setRightBarItem(oldItem, item, rightBarItem, false, Direction.LEFT);
        _layOutMembers();
	}

	public void setSinglePanel(Panel panel) {
        navStack.setSinglePanel(panel);
	}

	public ContainerFeatures getContainerFeatures() {
		return new ContainerFeatures(false, false, false, true, 0);
	}
}
